
/* Copyright, 1999-2000, QNX Software Systems Ltd.  All Rights Reserved

 * This source code has been published by QNX Software Systems
 * Ltd. (QSSL). However, any use, reproduction, modification, distribution
 * or transfer of this software, or any software which includes or is based
 * upon any of this code, is only permitted if expressly authorized by a
 * written license agreement from QSSL. Contact the QNX Developer's Network
 * or contact QSSL's legal department for more information.
 */


/*
 * egalaxy.c
 *
 *  eGalaxy  TouchKit serial controller  protocol (www.egalax.com.tw).
 *  ___ controller card, ___ touch screen
 *   Artem Smolenkov, 2004 
*	asmolenkov@yahoo.co.uk
 */

#include "devi.h"

#define	TRUE					1
#define	FALSE					!TRUE
#define FLAG_INIT				0x1000
#define FLAG_RESET				0x2000
#define DEBOUNCE			 	80000L


struct data {
	int					flags;
     struct 	packet_abs	                tp;       
	int					baud;     /* data transfer baud rate */
	int					state;    /* current state (see eg_input for details) */
	int					pktsize;
};

static int eg_init(input_module_t *module);
static int eg_devctrl(input_module_t *module,int event,void *ptr);
static int eg_reset(input_module_t *module);
static int eg_input(input_module_t *module, int num, void *ptr);
static int eg_parm(input_module_t *module,int opt,char *optarg);
static int eg_shutdown(input_module_t *module, int delay);


input_module_t	 egalaxy = {

	NULL,
	NULL,
	NULL,
	0,
	DEVI_CLASS_ABS|DEVI_MODULE_TYPE_PROTO,
	"egalaxy",
	__DATE__,
	"R",
	NULL,			// initialized by us (in init)
	eg_init,
	eg_reset,
	eg_input,
	NULL,
	NULL,
	eg_parm,
	eg_devctrl,
	eg_shutdown,
};


/* Description: callback initialisation function; it is called when input module    */
/*              initializes input system                                            */
/* Input      : input_module_t * module - pointer to module descriptor              */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise - (-1)                                           */
int eg_init(input_module_t *module)
{
        struct data *dp = module->data;
  
        if(!module->data)  {
                
                if(!(dp = module->data = calloc(sizeof *dp, 1))) {
                        return (-1);
                }
                dp->flags = ABSOLUTE | ABS_UNCALIBRATED | FLAG_RESET;
                dp->baud = 9600;
                dp->state = 0;
        }
        
        return (0);
}


/* Description: this is a callback function for DEVCTRL command processing          */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int event  - DEVCTRL command code                                   */
/*              void * ptr - pointer to data exchange block                         */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int eg_devctrl(input_module_t *module, int event, void *ptr)
{
	struct data 	*dp = module->data;
	input_module_t	*down = module->down;

	switch(event) {

		case DEVCTL_GETDEVFLAGS:
			*(unsigned short *)ptr = (dp->flags & FLAGS_GLOBAL);
			break;
		case DEVCTL_GETPTRBTNS:
			*(unsigned long *)ptr = 1L;
			break;
		case DEVCTL_GETPTRCOORD:
			*(unsigned char *)ptr = '\02';
			break;
	 	case DEVCTL_GETPTRPRESS:
			*(unsigned char *)ptr = 0;
			break;
	 	case DEVCTL_GETCOORDRNG: {
			struct devctl_coord_range	*range = ptr;
			range->min = 0;
			range->max = 2047;
			break;
		}
	 	case DEVCTL_GETPRESSRNG: {
			struct	devctl_coord_range	*range = ptr;
			range->min = 0;
			range->max = 0; // 255
			break;
		}
		default:
			if(down && down->devctrl) {
				return((down->devctrl)(down, event, ptr));
			} else {
				return (-1);
			}
	}

	return (0);
}

        
/* Description: this callback funtion is called when the module is linked into the  */
/*              event bus;it is used to set initial module state on the protocol    */
/*              level                                                               */
/* Input      : input_module_t * module - pointer to module descriptor              */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int eg_reset(input_module_t *module)
{
        struct data			*dp = module->data;
        input_module_t			*down = module->down;
        struct devctl_setbaud	        portctl;
        struct devctl_settrig	        trig;
        
        if((dp->flags & FLAG_INIT) == 0) {
                
                if(down && down->devctrl) {
                        
                        portctl.baud = dp->baud;
                        portctl.cflag = PNONE | STOP1 | CS8;
                        (down->devctrl)(down, DEVCTL_SETBAUD, &portctl);
                        
                        if(dp->flags & FLAG_RESET) {
                            (down->devctrl)(down, DEVCTL_RESET, NULL);
                            if (verbosity)
                               printf("penmount: reset fd\n");    
                        }
                        
                        trig.trigger_level = dp->pktsize = 1; // ? 5 ?
                        (down->devctrl)(down, DEVCTL_SETTRIG, &trig);
                        
                }
                
                dp->state = 0;
                dp->flags |= FLAG_INIT;
        }
        return (0);
}


/* Description: main protocol processing function.  It will be called  by the       */
/*              device layer to pass it data to process.  Its job is to interpret   */
/*              the data according to the PenMount touch screen protocol, create    */
/*              a data structure, fill it in and hand it off to the  filter layer   */
/*              module above it. The protocol is processed using simple state       */
/*              machine.  State info is kept in module's private data               */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int num - number of bytes to process                                */
/*              void * arg - raw data to process                                    */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int eg_input(input_module_t *module, int num, void *ptr)
{
        int					i, enqueue = 0;
        unsigned char			*p = (unsigned char *)ptr;
        struct data				*dp = module->data;
        input_module_t				*up = module->up;
        
        if(dp->flags & FLAG_INIT) {
                
                for (i = 0; i < num; i++, p++) {
                        
                     switch (dp->state) {   
                        case 0: 		// byte 0  :   1000 000St, St=1 for touch down St=0 for touch up
                                if ( *p & 0x80) {
                                        dp->state++;
                                        if ( *p & 0x01 ) 
										dp->tp.buttons = 4L;  //button pressed
								else
										dp->tp.buttons = 0;  //button pressed
                                }
                                break;
                        case 1: 		// byte 1: 0000 A10 A9 A8 A7
                               if (!(*p & 0xf0) ){
                                        dp->state++;
                                        dp->tp.x = *p;
                                        dp->tp.x <<= 7;
                                        dp->tp.x &= 0x0780;
                                    }else{
                                        dp->state=0;
                                }        
                                break;       
                        case 2:	// byte 2: 0 A6 A5 A4 A3 A2 A1 A0
                              if (!(*p & 0x80)){
                                      dp->state++;
                                      dp->tp.x |= *p;
                                  }else{
                                      dp->state=0;
                                }   
                                break; 
                        case 3:   // byte 3: 0000 B10 B9 B8 B7
                              if ( !(*p & 0xf0)){
                                     dp->state++;
                                     dp->tp.y = *p;
                                     dp->tp.y <<= 7;
                                     dp->tp.y &= 0x0780;
                                  }else{
                                     dp->state=0;
                                }
                                break;          
                        case 4: // byte 4: 0 B6 B5 B4 B3 B2 B1 B0
                              if (!(*p & 0x80)){
                                      dp->tp.y |= *p;
		                             dp->tp.z = 0;
		                             enqueue = 1;
                                  }
                              dp->state=0;
                              break; 
                           }
                        if (enqueue) {
                                dp->tp.flags = dp->flags;
                                clk_get(&dp->tp.timestamp);
                                enqueue = 0;
                                (up->input)(up, 1, &dp->tp);
                        }
                }
        }
        return (0);
}


/* Description: this is a callback function for command line parameter processing   */
/*                                                                                  */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int opt  - parameter code                                           */
/*              char * optarg - optional parameter value                            */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int eg_parm(input_module_t *module, int opt, char *optarg)
{
	struct data *dp = module->data;

 	switch (opt) {
		case 'R':
			dp->flags &= ~FLAG_RESET;
			break;
 		default:
 			return (-1);
 	}

	return (0);
}


/* Description: this is a callback function which is called when resourse manager   */
/*              is shutting down                                                    */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int delay  - program doesn't use this parameter                     */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
/* Comment    : Does nothing for the protocol level                                 */
int eg_shutdown(input_module_t *module, int delay)
{
	return (0);
}

