/*
 * Copyright 2001, QNX Software Systems Ltd. All Rights Reserved
 *
 * This source code has been published by QNX Software Systems Ltd. (QSSL).
 * However, any use, reproduction, modification, distribution or transfer of
 * this software, or any software which includes or is based upon any of this
 * code, is only permitted under the terms of the QNX Open Community License
 * version 1.0 (see licensing.qnx.com for details) or as otherwise expressly
 * authorized by a written license agreement from QSSL. For more information,
 * please email licensing@qnx.com.
 */

/*
 * elo.c
 *
 * Elographics touch screen  protocol. 
 *
 */

#include "devi.h"
#include <pthread.h>
#include <time.h>

#define	TRUE					1
#define	FALSE					!TRUE
#define DEBOUNCE			 	80000L

#define SMART_REPLY_SIZE                        (8)
#define SMART_CMD_SIZE                          (8)

#define MAX_TRIES                               (10)     /* Maximum number of tries to reset controller  */
#define TIME_WAIT                               (1)      /* Time interval (in seconds )  we wait until controller's response comes */

struct data {
#define FLAG_INIT				0x1000
#define FLAG_RESET				0x2000
	int					flags;
        struct packet_abs	                tp;       
	int					baud;     /* data transfer baud rate */
	int					state;    /* current state (see smart_input for details) */
	int					pktsize;
        /* next structure's fields we use to send command to SmartSet controller and to receive response */
        pthread_mutex_t                         mutex;
        pthread_cond_t                          cond;
        int                                     expect_reply_len; // Number of expected bytes (in reply mode)
        int                                     received_len;     // Number of received bytes (in reply mode)
        _uint8                                  reply_buffer[SMART_REPLY_SIZE + 2]; // Buffer to accept reply from 
};

static int smart_init(input_module_t *module);
static int smart_devctrl(input_module_t *module,int event,void *ptr);
static int smart_reset(input_module_t *module);
static int smart_input(input_module_t *module, int num, void *ptr);
static int smart_parm(input_module_t *module,int opt,char *optarg);
static int smart_shutdown(input_module_t *module, int delay);
static int send_out_string(_uint8 * inbuf, int nLen, input_module_t *module);
static void * reset_controller(void * parm);


input_module_t	smartset = {

	NULL,
	NULL,
	NULL,
	0,
	DEVI_CLASS_ABS|DEVI_MODULE_TYPE_PROTO, 
	"Smartset",
	__DATE__,
	"Rb:",
	NULL,			// initialized by us (in init)
	smart_init,
	smart_reset,
	smart_input,
	NULL,
	NULL,
	smart_parm,
	smart_devctrl,
	smart_shutdown, 
};


/* Description: callback initialisation function; it is called when input module    */
/*              initializes input system                                            */
/* Input      : input_module_t * module - pointer to module descriptor              */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise - (-1)                                           */
int smart_init(input_module_t *module)
{
        struct data *dp = module->data;
  
        if(!module->data)  {
                
                if(!(dp = module->data = calloc(sizeof *dp, 1))) {
                        return (-1); 
                }
                dp->flags = ABSOLUTE | ABS_UNCALIBRATED | FLAG_RESET;
                dp->baud = 9600;
                dp->state = 0;
        }
        
        return (0);
}
  

/* Description: this is a callback function for DEVCTRL command processing          */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int event  - DEVCTRL command code                                   */
/*              void * ptr - pointer to data exchange block                         */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int smart_devctrl(input_module_t *module, int event, void *ptr)
{
	struct data 	*dp = module->data;
	input_module_t	*down = module->down;

	switch(event) {

		case DEVCTL_GETDEVFLAGS:
			*(unsigned short *)ptr = (dp->flags & FLAGS_GLOBAL);
			break;
		case DEVCTL_GETPTRBTNS:
			*(unsigned long *)ptr = 1L;
			break; 
		case DEVCTL_GETPTRCOORD:
			*(unsigned char *)ptr = '\02';
			break;
	 	case DEVCTL_GETPTRPRESS:
			*(unsigned char *)ptr = (dp->flags & ABS_PRESS_DATA)?1:0;
			break;
	 	case DEVCTL_GETCOORDRNG: {
			struct devctl_coord_range	*range = ptr;

			// we don't know this, pass it on maybe the device knows ?
			if((down->devctrl)(down, DEVCTL_GETCOORDRNG, ptr) == -1) {
				// the lower layer doesn't know.  probably a serial device
				// well, let's try and guess.  If there is pressure data
				// then it must be 12 bit.  Otherwise we'll assume that
				// it's 8 bit
				if(dp->flags & ABS_PRESS_DATA) {
					range->min = 0;
					range->max = 4095;
				} else { 
					range->min = 0;
					range->max = 255;
				} 
			} 
			break;
		}
	 	case DEVCTL_GETPRESSRNG: {
			// for all smartset devices at present the range is 0-255
			struct	devctl_coord_range	*range = ptr;

			range->min = 0;
			range->max = 255;
			break;
		}
		default:
			if(down && down->devctrl) {
				return((down->devctrl)(down, event, ptr));
			} else {
				return (-1);
			}
	}

	return (0);
}


/* Description: this callback funtion is called when the module is linked into the  */
/*              event bus;it is used to set initial module state on the protocol    */
/*              level                                                               */
/* Input      : input_module_t * module - pointer to module descriptor              */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int smart_reset(input_module_t * pModule)
{
        struct data			*dp = pModule->data;
        input_module_t			*down = pModule->down;
        struct devctl_setbaud	        portctl;
        struct devctl_settrig	        trig;
        
        if((dp->flags & FLAG_INIT) == 0) {
                
                if(down && down->devctrl) {
                        
                        portctl.baud = dp->baud;
                        portctl.cflag = PNONE | STOP1 | CS8;
                        (down->devctrl)(down, DEVCTL_SETBAUD, &portctl);
                        
                        if(dp->flags & FLAG_RESET) {
                                (down->devctrl)(down, DEVCTL_RESET, NULL);
                        }
                        
                        trig.trigger_level = dp->pktsize = 1;
                        (down->devctrl)(down, DEVCTL_SETTRIG, &trig);
                        
                }
                
                dp->state = 0;
                dp->flags |= FLAG_INIT;
        }
        
        if(down && down->output) 
            {
	    // Start service thread to rset SmartSet controller end set its parameters
            pthread_attr_t attr;
            dp -> state = -1;
            pthread_attr_init(&attr);
	    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
            pthread_create(NULL, &attr, &reset_controller, pModule);
	    }
        
        return (0);
}

/* Description: this is a service function that prepare and send command to SmartSet*/
/*              controller                                                          */
/* Input      : _uint8 * inbuf - command to send                                    */
/*              int nLen - command length (in bytes)                                */
/*              input_module_t * module - pointer to module descriptor              */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
/* Description: Serial controller wait 10-bytes command:                            */
/*                -<Leading Byte> (usually 'U')                                     */
/*                -<Command Byte> (see documentation)                               */
/*                -<Command Parameters> (up to 7)                                   */
/*                -<Check Sum> (0xAA + 'U' + <Cmd> + all <Cmd Parms>)               */
int send_out_string(_uint8 * inbuf, int nLen, input_module_t * pModule)
{
   _uint8 outbuf[SMART_CMD_SIZE + 2];
   _uint8 * pin  = inbuf, 
         * pout  = outbuf;
   int i = 0;
   _uint8 sum;
   struct data	*dp = pModule -> data;
   input_module_t * down = pModule->down;
   struct timespec tw;
   int rc = EOK;
   
   *pout = 'U'; // Leading byte
   sum = 0xaa + *pout;
   ++pout;
   for(i = 0; i < nLen; ++i)
     {
     *pout = *pin;
     sum += *pin;
     ++pin; ++pout;
     }
   for(; i < SMART_CMD_SIZE; ++i) // Pad with zeroes
     {
     *pout = 0;
     ++pout;
     }
   *pout = sum; // Check sum

   // Prepare reception
   dp -> expect_reply_len = SMART_REPLY_SIZE + 2;
   dp -> received_len = 0;
   // Send it
   if(verbosity >= 5)
       {
       printf("Command sent : ");
       for(i = 0; i < (SMART_CMD_SIZE + 2); ++i)
           printf("%#x ", (int)outbuf[i]);
       printf("\n");
       }
   
   tw.tv_sec  = time(0) + TIME_WAIT;
   tw.tv_nsec = 0;
   
   pthread_mutex_lock(&(dp -> mutex));
   (down->output)(down, outbuf, SMART_CMD_SIZE + 2);
   // Wait for response
   rc = pthread_cond_timedwait(&(dp -> cond), &(dp -> mutex), &tw);
   
   if(EOK != rc)
     {
     if(verbosity)
	  fprintf(stderr, "SmartSet controller doesn't resonse (error code = %i)\n", rc);
     return (-1);
     }
//   while(dp -> expect_reply_len >  dp -> received_len);
   // Analize resonse
   return (((dp -> reply_buffer[0] == 'U') && 
   	    (dp -> reply_buffer[1] == 'A') &&
	    (dp -> reply_buffer[2] == '0')) ? 0 : -1);

}


/* Description: this is a service thread function that reset  SmartSet controller   */
/*              it runs in separate thread to wait controller response              */
/* Input      : void * parm - really it is   input_module_t * module - pointer to   */
/*                            module descriptor                                     */
/* Output     : None                                                                */
/* Return     : None                                                                */
void * reset_controller(void * parm)
{
   static _uint8 reset[] = {'R', 0x01 };  // Soft controller reset
   static _uint8 mode[] =  {'M', 0, 0x06, 0x40}; // Set controller mode: STREAM ON, UNTOUCH ON, TRACKING ON
   input_module_t * pModule = (input_module_t *)parm;
   struct data	*dp = pModule -> data;
   int i;
   char * pMsgTxt = "Error: %s - unable to reset touch screen controller\n";
   
   pthread_mutex_init(&(dp -> mutex), NULL);
   pthread_cond_init(&(dp -> cond), NULL);

   for(i = 0; i < MAX_TRIES; ++i)
     {
     if((0 == send_out_string(reset, sizeof(reset), pModule)) && // Reset
        (0 == send_out_string(mode, sizeof(mode), pModule)))     // Set mode
	{
        pthread_mutex_destroy(&(dp -> mutex));
        pthread_cond_destroy(&(dp -> mutex));
        dp -> state = 0; // Ready to get income data
	return 0;
	}
     }

  fprintf(stderr, pMsgTxt, pModule -> name); 
  slogf(_SLOG_SETCODE(_SLOGC_INPUT, 0), _SLOG_ERROR, pMsgTxt, pModule -> name);
  exit(-1);

}

/* Description: main protocol processing function.  It will be called  by the       */
/*              device layer to pass it data to process.  Its job is to interpret   */
/*              the data according to the Elographics touch screen protocol, create */
/*              a data structure, fill it in and hand it off to the  filter layer   */
/*              module above it. The protocol is processed using simple state       */
/*              machine.  State info is kept in module's private data               */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int num - number of bytes to process                                */
/*              void * arg - raw data to process                                    */
/* Output     : None                                                                */ 
/* Return     : 0 if OK, otherwise -1                                               */
int smart_input(input_module_t *module, int num, void *ptr) 
{
        int					i, enqueue = 0;
        char					*p = (char *)ptr;
        struct data				*dp = module->data;
        input_module_t				*up = module->up;
        
        if(verbosity >= 5)
           {
           printf("Received : ");
           for(i = 0; i < num; ++i)
               printf("%#x ", (int)*((_uint8 *)ptr + i));
           printf("\n");
	   }
   
        if(dp->flags & FLAG_INIT) {  
                
                for (i = 0; i < num; i++, p++) {
                        
                        switch (dp->state) {
			case -1: // Waiting for responce
			   pthread_mutex_lock(&(dp -> mutex));
			   if(dp -> received_len <  sizeof(dp -> reply_buffer))
			        {
			        dp -> reply_buffer[dp -> received_len] = *p;
				dp -> received_len++;
				}
			   if(dp -> expect_reply_len <=  dp -> received_len)
			        {
				pthread_cond_signal(&dp -> cond);
			        pthread_mutex_unlock(&(dp -> mutex));
			        }
                           break;
                        case 0: 
                                // Leading byte 
                                if (*p == 'U')
                                        dp->state = 1;
                                break;
                                
                        case 1:
                                // Means this is a touch message 
                                if(*p == 'T') {
                                        dp->state = 2;
                                } else {
                                        if (verbosity >= 3) 
                                               fprintf(stderr, "SmartSet: unknown CMD BYTE (%#x)\n", (int)*p);
                                        dp->state = 0;
                                }
                                break;
                                
                        case 2:
                                // Button :-)) state
                                if(*p & 0x04) {
                                        dp->tp.buttons = 0;
                                } else {
                                        dp->tp.buttons = 4L;
                                }
                                dp->state++;
                                break; 
                                
                        case 3: 
                                // X-coordinate
                                dp->tp.x = (unsigned char ) *p;
                                dp->state++;
                                break;
                                
                        case 4:
                                // X-coordinate
                                dp->tp.x |= ((unsigned char)*p << 8);
                                dp->state++;
                                break;
                                
                        case 5:
                                
                                // Y-coordinate
                                dp->tp.y = (unsigned char)*p;
                                dp->state++;
                                break;
                                
                        case 6: 
                                
                                // Y-coordinate
                                dp->tp.y |= ((unsigned char)*p << 8);
                                dp->state++;
                                break;
                                
                        case 7:
                                
                                // Z-coordinate
                                dp->tp.z = ((unsigned char)*p << 8);
                                dp->state++;
                                break;
                                
                        case 8:
                                
                                // Z-coordinate
                                dp->tp.z |= (unsigned char)*p;
                                dp->state++;
                                break;
                                
                        case 9: 
                                // Nothing
			        dp->state = 0;
                                enqueue = 1;
                                break;
                        } 
                        
                        if (enqueue) {
                                
                                dp->tp.flags = dp->flags;
                                clk_get(&dp->tp.timestamp);
                                enqueue = 0;
                                
                                (up->input)(up, 1, &dp->tp);
                        }
                }
        }  
        return (0);
}


/* Description: this is a callback function for command line parameter processing   */
/*              (all valid parameters for device module are listed in smartset.args)*/
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int opt  - parameter code                                           */
/*              char * optarg - optional parameter value                            */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int smart_parm(input_module_t *module, int opt, char *optarg)
{
	struct data *dp = module->data;

 	switch (opt) {
		case 'R':
			dp->flags &= ~FLAG_RESET;
			break;
		case 'b':
			dp->baud = atoi(optarg);
			break;
 		default:
 			return (-1);
 	}

	return (0);
}


/* Description: this is a callback function which is called when resourse manager   */
/*              is shutting down                                                    */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int delay  - program doesn't use this parameter                     */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
/* Comment    : Does nothing for the protocol level                                 */
int smart_shutdown(input_module_t *module, int delay)
{
	return (0);
}

