/* 
 * Copyright 2000-2001, QNX Software Systems Ltd.  All Rights Reserved
 *
 * This source code has been published by QNX Software Systems Ltd.
 * (QSSL).  However, any use, reproduction, modification, distribution
 * or transfer of this software, or any software which includes or is
 * based upon any of this code, is only permitted under the terms of
 * the QNX Open Community License version 1.0 (see licensing.qnx.com
 * for details) or as otherwise expressly authorized by a written license
 * agreement from QSSL.  For more information, please email licensing@qnx.com.
 */

/*
 * kbd.c
 *
 * Keyboard protocol module.
 *
 */

#include "devi.h"
#include "pthread.h"


/* forward declarations                                                                               */
static int kbd_init(input_module_t *module);
static int kbd_devctrl(input_module_t *module, int event, void *ptr);
static int kbd_reset(input_module_t *module);
static int kbd_input(input_module_t *module, int num, void *arg);
static int kbd_parm(input_module_t *module, int opt, char *optarg);
static int kbd_shutdown(input_module_t *module, int delay);

/* Our protocol module is represented by the following input_module_t  data structure */

input_module_t	kbd = {
	NULL,
	NULL,
	NULL,
	0,
	DEVI_CLASS_KBD|DEVI_MODULE_TYPE_PROTO,  
	"Kbd",
	__DATE__,
	"srRk:f:p:",
	NULL,			              /* initialized by us (in init)       */
	kbd_init,
	kbd_reset,
	kbd_input,
	NULL,
	NULL,
	kbd_parm,
	kbd_devctrl,
	kbd_shutdown
};

#define FLAG_INIT			0x1000
#define FLAG_RESET			0x2000 /* Do reset of device level            */
#define FLAG_SYM                        0x8000 /* DEvice level supplies valid symbols */

#define MAX_SCAN_LEN                    (10)   // Maximum number of scan codes per one USB code 

/* Protocol module private data                                                       */
struct data {
	int			flags;
	int			rate;   /* repeat rate                                */
	int		        delay;  /* repeat delay                               */
        int                     fId;    /* For debug mode (-f parameter) - write input to file  */
        int                     fifo;   /* For debug mode (-p parameter) - write input to FIFO file  */
        char                    fifo_name[100];
        int                     nScanInBuf;
        unsigned char           scanBuf[MAX_SCAN_LEN];
};

void *  open_pipe(void * parm)
{
     struct data * dp = (struct data *)parm;
     int pipe_no;
   
     dp -> fifo = 0;
     // Block until O_RDONLY pipe will be opened
     pipe_no = open(dp -> fifo_name, O_WRONLY );
     close(pipe_no);
     // Wait a peer process re-opens pipe
     sleep(1);
     // Open non-blocked pipe
     dp -> fifo =  open(dp -> fifo_name, O_WRONLY | O_NONBLOCK );
   
     return 0;
}


/* Description: function returns index of valid keyboard rate                       */
/* Input      : int rate - rate value                                               */
/* Output     : None                                                                */
/* Return     : index that corrsponds to rate                                       */
static int validate_rate(int rate)
{
	int					i;
	static const int	valid_rates[] = {
		300, 267, 240, 218, 200, 185, 171, 160,
		150, 133, 120, 109, 100,  92,  86,  80,
		 75,  67,  60,  55,  50,  46,  43,  40,
		 37,  33,  30,  27,  25,  23,  21,  20
	};

	for(i = 0; i < sizeof valid_rates / sizeof *valid_rates; i++) {
		if(rate >= valid_rates[i]) {
			break;
		}
	}

	if(i < sizeof valid_rates/sizeof *valid_rates) {
		return valid_rates[i];
	} else {
		return -1;
	}
}


/* Description: callback initialisation function; it is called when input module is */
/*              initialising the input system                                       */
/* Input      : input_module_t * module - pointer to module descriptor              */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise - (-1)                                           */
int kbd_init(input_module_t	*module)
{
	struct data *dp=module->data;

 	if(!module->data)  {
 		if(!(dp = module->data = calloc(sizeof *dp, 1))) {
 			return -1;
 		}
		dp->flags |= FLAG_RESET;
 		dp->rate=300;   /* in times per second multiply by 10                    */
		dp->delay=1;    /* In relative units (see devctl) - corresponds 250 msec */
		dp->fId = 0;
		dp -> fifo = 0;
                signal(SIGPIPE, SIG_IGN);	   
 	}

	return(0);
}

/* Description: this is a callback function for DEVCTRL command processing          */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int event  - DEVCTRL command code                                   */
/*              void * ptr - pointer to data exchange block                         */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int kbd_devctrl(input_module_t *module,int event,void *ptr)
{
	struct data *dp=module->data;
	input_module_t	*down=module->down;

	switch(event) {
		case DEVCTL_RESET_KBD:
			if(dp->flags & FLAG_RESET) {
				int led=0;

				(down->devctrl)(down, DEVCTL_LED, &led);
			}
			break;
		case DEVCTL_RESET_PS2MOUSE:
			break;
		case DEVCTL_SETKBD: {
			struct devctl_setkbd	*devctl=ptr;
			int    rate;

			devctl->delay = (devctl->delay + 249U) / 250U - 1U;
			devctl->delay = min(max(devctl->delay, 0), 3);
			if((rate=validate_rate(devctl->rate)) == -1) {
				return -1;			
			} else {
				devctl->rate=rate;
			}

			if(down && down->devctrl) {
				(down->devctrl)(down, DEVCTL_SETKBD, devctl);
			}
			
			break;			
		}
		default:
			if(down && down->devctrl) {
				return((down->devctrl)(down, event, ptr));
			} else {
				return -1;
			}
	}

	return EOK;
}

extern input_module_t keyboard_filter;

/* Description: this callback funtion is called when the module is linked into the  */
/*              event bus; it is used to set initial module state on the protocol   */
/*              level                                                               */
/* Input      : input_module_t * module - pointer to module descriptor              */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int kbd_reset(input_module_t *module)
{
        struct data 			*dp = module->data;
        input_module_t			*down = module->down;
  
	keyboard_filter.reset(&keyboard_filter);
        if((dp->flags & FLAG_INIT) == 0) {

                if(down && down->devctrl) {
                        
                        if(dp->flags & FLAG_RESET && dp->rate != -1) {
                                struct devctl_setkbd	devctl;
	
                                devctl.rate = dp->rate;
                                devctl.delay = dp->delay;
                                (down->devctrl)(down, DEVCTL_SETKBD, &devctl);
                        }
      
                }
                dp->flags |= FLAG_INIT;
        }
        dp -> nScanInBuf = 0;
        return (0);
}

/* Description: main protocol processing function.  It will be called  by the       */
/*              device layer to pass it data to process.  Its job is very simple:   */
/*              just to hand data off to the  filter layer  module above it.        */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int num - number of bytes to process                                */
/*              void * arg - raw data to process                                    */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
/* Comment    : driver uses one of two input modules depending on device data       */
/*              protocol                                                            */  
int kbd_input(input_module_t *module, int num, void *ptr)
{
	struct data				*dp=module->data;
	input_module_t			*up=module->up;
	struct packet_kbd		kbuf;
	unsigned char					*tmp=ptr;
	int				i;
        int nLen;

        if(dp->flags & FLAG_INIT) {


		for(i=0; i<num; i++) {

			memset(&kbuf, 0x00, sizeof kbuf);
                        
                        if (dp->flags & FLAG_SYM) {  
			   /* The device level supplies valid symbls                */
				kbuf.key.key_sym = *tmp++;
				kbuf.key.flags = KEY_SYM_VALID;
                        }
                        else {
	                        if(0 < dp -> fId)
	                            write(dp -> fId,  tmp,  1);
			   
	                        if(0 < dp -> fifo)
			            {
	                            if(0 > write(dp -> fifo,  tmp,  1))
					{
					dp -> fifo = 0;
   	                                pthread_create(NULL, NULL, open_pipe, dp);
					}
				    }
			   /* Th device level supplies scan codes                   */
			       
                                kbuf.key.key_scan = *tmp++;
				kbuf.key.flags = KEY_SCAN_VALID;
			        nLen = 1;
			        // Collect scan codes in temporary buffer until scan codes sequence
			        // can be converted to USB code
			        dp -> scanBuf[dp -> nScanInBuf] = kbuf.key.key_scan;
				++(dp -> nScanInBuf);
			        if((EOK != devi_scan_to_usb(dp -> scanBuf, 
							   &(dp -> nScanInBuf),
							   &kbuf.nUSBCode,
							   &kbuf.bMakeBreak,
							   &nLen)) ||
				   (0== nLen))
			           {
			           kbuf.nUSBCode = KS_None;
				   if(dp -> nScanInBuf == MAX_SCAN_LEN) // Too long! Maybe keyboard error. Reset
				       kbd_reset(module); 	
				   } 
			        // Otherwise devi_scan_to_usb removes "used" scan codes from the top of  
				// scanBuf and subtracts number of removed scan codes from nScanInBuf
			           
			}
			clk_get(&kbuf.timestamp);
			(up->input)(up, 1, &kbuf);
		}
	}

	return(0);
}

/* Description: this is a callback function for command line parameter processing   */
/*              (all valid parameters for device module are listed in kbd.args)     */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int opt  - parameter code                                           */
/*              char * optarg - optional parameter value                            */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int kbd_parm(input_module_t *module,int opt,char *optarg)
{
	struct data	*dp=module->data;

	switch(opt) {
		case 'r':    /* reset device level while doing protocol reset       */
			dp->flags |= FLAG_RESET;
			break;
		case 'R':    /* don't reset evice level while doing protocol reset  */
			dp->flags &= ~FLAG_RESET;
			break;

                case 's':    /* Device level should supply valid symbols            */
                        dp->flags &= FLAG_SYM;
                        break;
		case 'k':    /* specify key repeat rate and delay                   */
			dp->rate = strtol(optarg, (char **)&optarg, 10) * 10;
			if(*optarg == '.' && *++optarg >= '0' && *optarg <= '9') {
				dp->rate += *optarg++ - '0';
			}
			if(*optarg == ',') {
				dp->delay = (strtol(++optarg, 0, 10) + 249U) / 250U - 1U;
				dp->delay = min(max(dp->delay, 0), 3);
			}
			if(dp->rate) {
				if((dp->rate=validate_rate(dp->rate)) == -1) {
	                                char * pMsgTxt = "Warning: invalid keyboard rate requested (%s)\n";
	                                fprintf(stderr, pMsgTxt, module->name);
	                                slogf(_SLOG_SETCODE(_SLOGC_INPUT, 0), _SLOG_WARNING, pMsgTxt, module->name);
				}
			} else {
				dp->rate = -1;
			}
			break;
     	        case 'f':
	                dp -> fId = open(optarg, O_CREAT | O_WRONLY | O_TRUNC,
				                        S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
	                break;
     	        case 'p':
	                if(access(optarg, F_OK))
			   {
			   if(mkfifo(optarg, S_IRUSR | S_IWUSR))
				{
				printf("Cannot open FIFO (%s), error = %i\n", optarg, errno); 
				break;
				}
			   printf ("Pipe %s has been created\n", optarg);
			   }
	                strcpy(dp -> fifo_name, optarg);
   	                pthread_create(NULL, NULL, open_pipe, dp);
	                break;
		default:
			return -1;
	}

	return(0);
}


/* Description: this is a callback function which is called when resourse manager   */
/*              is shutting down                                                    */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int ms  - program doesn't use this parameter                        */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
/* Comment    : Does nothing for the keyboard protocol level                        */
int kbd_shutdown(input_module_t *module, int delay)
{
    struct data *dp = module->data;
    if(0 < dp -> fId) /* if debug output file was created - close it */
       close(dp -> fId);
    if(0 < dp -> fifo) /* if debug output file was created - close it */
       close(dp -> fifo);
	return(0);
}

