/*
 * Copyright  2000, QNX Software Systems Ltd.  All Rights Reserved
 *
 * This source code has been published by QNX Software Systems Ltd.
 * (QSSL).  However, any use, reproduction, modification, distribution
 * or transfer of this software, or any software which includes or is
 * based upon any of this code, is only permitted under the terms of
 * the QNX Open Community License version 1.0 (see licensing.qnx.com
 * for details) or as otherwise expressly authorized by a written license
 * agreement from QSSL.  For more information, please email licensing@qnx.com.
 */


/*
 * This is a dvice module for keyboard support
 *
 */

#include "devi.h"
#include "ps2api.h"

/* module specific flags */

#define FLAG_INIT                0x00001  /* module has been initialized */
#define FLAG_RESET_KB            0x00008  /* reset the kb */
#define FLAG_INB_FULL            0x00010  /* irq couldn't add to input buff */

#define TRUE                     1
#define FALSE                    0

/* keyboard manifests */

#define ACK                      0xfa /* ack code */
#define RESEND                   0xfe /* resend */


#define KBD_IRQ                  1    /* controller default irq */
#define KBD_DEF_PORT             0x60 /* controller default ioport */
#define STATUS_PORT_ADD          0x04 /* number to add to i/o port to get status port number */


#define KBD_RESET                0xff /* reset the kbd */
#define KBD_RATE                 0xf3 /* set the repetition rate */
#define KBD_LEDS                 0xed /* set the leds (kbd lights) */

#define KBD_DELAY_250MS		 0x00 /* 250ms before repeat starts */
#define KBD_DELAY_500MS          0x01 /* 500ms */

#define KBD_BUFSIZE 32		/* size of the kbd input buffer */

/* global flags */


/* global flags                                           */
#define GLOB_KBD_CMD_BUSY         (1 << 0)
#define GLOB_KBD_EXP_ACK          (1 << 1)
#define GLOB_KBD_DRAINED          (1 << 3)
#define GLOB_KBD_RESET            (1 << 4)
#define GLOB_KBD_HAVE_CMD_BYTE    (1 << 5)


// Logical port numbers
#define KBD_PORT                   (0)


struct data {
        unsigned 		flags;
        unsigned                irq;
        unsigned                ioport;
        unsigned                statport;
        int                     priority; /* processing thread priority                         */
        int			rate;   /* keyboard sample rate                                 */
        int			delay;  /* repeat delay                                         */
        int                     fId;    /* For debug mode (-d parameter) - write input to file  */
};

  

/* forward decl */
static int  index_rate(int rate);
static void set_kbd_delay_rate(struct data *dp, int newDelayRate);
static void set_kbd_leds(struct data *dp, short leds);
static int  kbddev_init(input_module_t *module);
static int  kbddev_reset(input_module_t *module);
static int  kbddev_pulse(message_context_t *, int, unsigned, void *);
static int  kbddev_devctrl(input_module_t *module, int event, void *ptr);
static int  kbddev_parm(input_module_t *module, int opt, char *optarg);
static int  kbddev_shutdown(input_module_t *module, int ms);


/* The following input_module_t data structure represents our device   */
/* module in                                                          */ 

input_module_t	kbddev = {
	NULL,
	NULL,
	NULL,
	0,
	DEVI_CLASS_NONE|DEVI_MODULE_TYPE_DEVICE,
	"kbddev",
	__DATE__,
	"P:rp:i:b:f:",
	NULL,		/* initialized by us (in init) */
	kbddev_init,
	kbddev_reset,
	NULL,		/* this moduleulator is input & EVBL_TERMINATE */
	NULL,
        kbddev_pulse,
	kbddev_parm,
	kbddev_devctrl,
	kbddev_shutdown,
};

/* global flags */



/* Description: function returns index of valid keyboard rate                       */
/* Input      : int rate - rate value                                               */
/* Output     : None                                                                */
/* Return     : index that corrsponds to rate                                       */
int index_rate(int rate)
{
        int			i;
        static const int	valid_rates[] = {
                300, 267, 240, 218, 200, 185, 171, 160,
                150, 133, 120, 109, 100,  92,  86,  80,
                75,  67,  60,  55,  50,  46,  43,  40,
                37,  33,  30,  27,  25,  23,  21,  20
                };
        
        for(i = 0; i < sizeof valid_rates / sizeof *valid_rates; i++) {
                if(rate >= valid_rates[i]) {
                        break;
                }
        }
        return (i);
}




/* Description: function initialize keyboard device module                          */
/* Input      : input_module_t * module - pointer to module descriptor              */
/* Output     : None                                                                */
/* Return     : 0 if OK, othervise -1                                               */
int kbddev_init(input_module_t *module)
{
    struct data *dp = module->data;

   if(!module->data) 
     {

      if(!(dp = module->data = calloc(sizeof *dp,  1))) 
        return (-1);

      dp->irq = KBD_IRQ;
      dp->ioport = KBD_DEF_PORT;
      dp->statport = dp->ioport + STATUS_PORT_ADD;
	
      dp -> priority = 0; /* Use default */
       
      dp->rate = 300;
      dp->delay = KBD_DELAY_500MS;
	
      return 0;
     }
   return (-1);
}



/* Description: this callback funtion is called when the module is linked into the  */
/*              event bus;it is used to set initial module state on the device level*/
/* Input      : input_module_t * module - pointer to module descriptor              */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int kbddev_reset(input_module_t *module)
{
   struct data	*dp   = module->data;

   if((dp->flags & FLAG_INIT) == 0) 
     {

     if(init_contr(dp->flags & FLAG_RESET_KB ? INIT_SELF_TEST : 0,
		   dp -> irq,
		   SET_NONE, 
		   dp -> ioport, dp -> statport))
	  {
          char * pMsgTxt = "Error: PS/2 controller initialization error\n. The driver is terminating";
          fprintf(stderr, pMsgTxt);
          slogf(_SLOG_SETCODE(_SLOGC_INPUT, 0), _SLOG_WARNING, pMsgTxt);
	  return -1;
	  }

     if(register_device(DEFAULT_DISPHANDLE, kbddev_pulse, DEFAULT_COID, KBD_PORT, module))
	  return -1;
     set_priority(KBD_PORT, dp -> priority);
     dp->flags |= FLAG_INIT;
     }
   /* setup appropriate device */

   reset_device(KBD_PORT);
   send_command(KBD_PORT, KBD_RESET, NULL, 0);
   set_kbd_delay_rate(dp, (dp->delay & 0x03) << 5 | index_rate(dp->rate));
   send_command(KBD_PORT, 0xf4, NULL, 0);
   return (0);
}


/* Description: function sets keyboard delay and rate                               */
/* Input      : struct data *dp - pointer to common driver data block               */
/*              int newRate - new sample rate                                       */
/* Output     : None                                                                */
/* Return     : None                                                                */
void set_kbd_delay_rate(struct data *dp, int newDelayRate)
{
   unsigned char setRateCmds[] =
     {
     KBD_RATE, 0
     };

   if(verbosity >= 3)
     printf("Setting delay/rate; command byte = %#x \n", newDelayRate);
   setRateCmds[1] = newDelayRate;
   send_command_set(KBD_PORT, setRateCmds, sizeof(setRateCmds), NULL, 0);
   
}

/* Description: function sets keyboard delay and rate                               */
/* Input      : struct data *dp - pointer to common driver data block               */
/*              int leds - led state byte                                           */
/* Output     : None                                                                */
/* Return     : None                                                                */
void set_kbd_leds(struct data *dp, short leds)
{
   unsigned char setLedsCmds[] =
     {
     KBD_LEDS, 0   
     };

   setLedsCmds[1] = leds & (KEYIND_SCROLL_LOCK | KEYIND_NUM_LOCK | KEYIND_CAPS_LOCK);
   send_command_set(KBD_PORT, setLedsCmds, sizeof(setLedsCmds), NULL, 0);
}

/* Description: this is a callback that kernel activate asychroniously each time    */
/*              when keyboard or mouse interrupt occures                            */
/* Input      : message_context_t *ctp - pointer to module descriptor               */
/*              int code - program doesn't use this parameter                       */
/*              unsigned flags - program doesn't use this parameter                 */
/*              void *data - pointer to (input_module_t	kb) structure               */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int kbddev_pulse(message_context_t *ctp, int code, unsigned flags, void *data)
{
    input_module_t		*module = (input_module_t *) data;
    struct data 		*dp = module->data;
    input_module_t 		*up = module->up;
    unsigned char 		c = get_in_data(ctp);
        
    if (verbosity >= 5 )
       printf("\tKeyboard device driver Received this byte %x\n", c);
    
    (up->input)(up, 1, (void *) &c);
    if(0 < dp -> fId)
        write(dp -> fId, &c, 1); 
  
    return (0);
}
    

/* Description: this is a callback function for DEVCTRL command processing          */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int event  - DEVCTRL command code                                   */
/*              void * ptr - pointer to data exchange block                         */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int kbddev_devctrl(input_module_t *module, int event, void *ptr)
{ 
    struct data *dp = module->data;

    switch(event) 
        {

        case DEVCTL_GETKBD: 
	    if(NULL != ptr)
	      { /* Get keyboard parameters - delay & rate */
              struct devctl_getkbd	*devctl = ptr;
                
              if(devctl) 
		 {
                 if(verbosity >= 8) 
                     printf("%s: responding with kb ctrl data [rate %d, delay %d]\n", 
                             module->name, dp->rate, dp->delay);
                        
                 devctl->rate = dp->rate;
                 devctl->delay = dp->delay;
                 }
              }
           break;
                
        case DEVCTL_SETKBD: 
	   if(NULL != ptr)
	     {		/* Set KB rate and delay */
             struct devctl_setkbd			*devctl = ptr;
                
             if(devctl) 
		{
                dp->rate = devctl->rate;
                dp->delay = devctl->delay;
                set_kbd_delay_rate(dp, 
		    (devctl->delay & 0x03) << 5 | 
                    index_rate(devctl->rate));
                        
                if(verbosity >= 4) 
                    printf("%s: setting kb ctrl parms [rate %d, delay %d]\n", module->name,
                    devctl->rate, devctl->delay);
                }
              }
                
            break;
        case DEVCTL_LED: 
	   if(NULL != ptr)
	     set_kbd_leds(dp, *(short *)ptr);
	   else
	     set_kbd_leds(dp, 0);
		
           break;
                
        case DEVCTL_RESET_KBD: 
                
           /* do nothing here, legacy */
           break;

        default:
           return (-1);
        }

    return (EOK);
}


/* Description: this is a callback function for command line parameter processing   */
/*              (all valid parameters for device module are listed in kb.args)      */ 
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int opt  - parameter cde                                            */
/*              char * optarg - optional parameter value                            */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int kbddev_parm(input_module_t *module, int opt, char *optarg)
{
        struct data *dp = module->data;


        switch (opt) {

        case 'r':
                dp->flags |= FLAG_RESET_KB;
                break;
                
        case 'i':
                dp->irq = strtoul(optarg, 0, 0);
                break;
                
        case 'p':
	        if(NULL != optarg)
	           {
		   char * p = strchr(optarg, ',');
                   dp->ioport = strtoul(optarg, 0, 16);
		   if(NULL != p)
		      dp->statport = dp->ioport  + strtol(p + 1, 0, 10);
		   else
		      dp->statport = dp->ioport  + STATUS_PORT_ADD;
		   }
                break;
	case 'P':
	       if(optarg)
	          {
		  int nPrty = atoi(optarg);
		  if(nPrty > 0) 
	             dp -> priority = nPrty;
		  }
                break;
                
   	case 'f':
	        dp -> fId = open(optarg, O_CREAT | O_WRONLY | O_TRUNC,
				 S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
	        break;
        default:
                return (-1);
        }
        
        return (0);
}

/* Description: this is a callback function which is called when resourse manager   */
/*              is shutting down                                                    */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int ms  - program doesn't use this parameter                        */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int kbddev_shutdown(input_module_t *module, int ms)
{
    struct data *dp = module->data;
 
    if (verbosity) 
	 printf("Disabling the keyboard\n");
    
    unregister_device(KBD_PORT);
   
    if(0 < dp -> fId) /* if debug output file was created - close it */
       close(dp -> fId);
   
    return (0);

}





