/* 
 * Copyright  2000, QNX Software Systems Ltd.  All Rights Reserved
 *
 * This source code has been published by QNX Software Systems Ltd. 
 * (QSSL).  However, any use, reproduction, modification, distribution
 * or transfer of this software, or any software which includes or is
 * based upon any of this code, is only permitted under the terms of
 * the QNX Open Community License version 1.0 (see licensing.qnx.com 
 * for details) or as otherwise expressly authorized by a written license
 * agreement from QSSL.  For more information, please email licensing@qnx.com.
 */
  
  
/*
 * This is a dvice module for PS2 mouse support 
 */
#include "devi.h"
#include "ps2api.h"

/* module specific flags */ 

#define FLAG_INIT                0x00001  /* module has been initialized */
#define FLAG_AUX                 0x00004  /* this module talks to the aux */
#define FLAG_RESET_KB            0x00008  /* reset the kb */

#define TRUE                     1
#define FALSE                    0

#define NOT_DEFINED              (-1)

/* keyboard manifests */

#define KBD_DEF_PORT             0x60 /* controller default ioport */
#define STATUS_PORT_ADD          0x04 /* number to add to i/o port to get status port number */


/* aux device manifests */

#define AUX_IRQ                  12   /* default aux irq */


#define AUX_INTR_ENABLE          0x47 /* cmd byte enabling aux ints */
#define AUX_ENABLE               0xa8 /* enable the aux device */
#define AUX_DISABLE              0xa7 /* disable the aux device */
#define AUX_ENABLE_XFER          0xf4 /* enable stream mode transfer */
#define AUX_DISABLE_XFER         0xf5 /* disable stream mode transfer */
#define AUX_SAMPLE_RATE          0xf3 /* set the sample rate */
#define AUX_200_SMP_SEC          0xC8 /* 200 samples/second */
#define AUX_100_SMP_SEC          0x64 /* 100 samples/second */
#define AUX_80_SMP_SEC           0x50 /* 80 samples/second  */
#define AUX_60_SMP_SEC           0x3c /* 60 samples/second  */
#define AUX_40_SMP_SEC           0x28 /* 40 samples/second  */
#define AUX_20_SMP_SEC           0x14 /* 20 samples/second  */
#define AUX_RESOLUTION           0xe8 /* set the resolution */
#define AUX_8_CNTS_MM            0x03 /* 8 counts/mm        */
#define AUX_4_CNTS_MM            0x02 /* 4 counts/mm        */
#define AUX_2_CNTS_MM            0x01 /* 2 counts/mm        */
#define AUX_1_CNTS_MM            0x00 /* 1 counts/mm        */
#define AUX_RESET                0xFF /* RESET mouse        */
#define AUX_COMPLETE             0xaa /* Mouse complete     */
#define AUX_GET_STATUS           0xe9 /* get mouse state    */
#define AUX_DEV_ID               0xf2 /* get device ID      */
#define AUX_STREAM_MODE          0xea /* Set stream mode    */


// Mouse replies
#define RESEND                   (0xfe) /* resend                                  */


// Logical port numbers
#define AUX_PORT                   (1)


#define NUM_OF_RETRIES             (5)


struct data {

        unsigned 		flags;
        unsigned                irq;
        unsigned                ioport;
        unsigned                statport;
        int                     priority; /* processing thread priority                         */
        int                     type;   /* 0 - regular, 3 - 3-buttons+wheel, 4 -4-buttons+wheel */ 
        int                     curtype;/* curent mouse mode                                    */
        int                     auxRate;/* mouse sample rate                                    */
        int                     auxResolution; /* mouse resolution                              */
        unsigned char           lastByte; /* previous byte (for mouse)                          */
        int                     fId;    /* For debug mode (-d parameter) - write input to file  */
};

typedef struct _mouse_state
{
   unsigned char rbPressed : 1; /* 1 - right button pressed       */    
   unsigned char mbPressed : 1; /* 1 - middle button pressed      */
   unsigned char lbPressed : 1; /* 1 - left button pressed        */
   unsigned char reserve1  : 1; /* reserve                        */
   unsigned char scaling   : 1; /* scaling -  1=2:1; 0=1:1        */
   unsigned char enabled   : 1; /* Enabled                        */
   unsigned char mode      : 1; /* 1 - remote, 0 - stream         */
   unsigned char reserve2  : 1; /* reserve                        */
   unsigned char resolution;    /* Resolution                     */
   unsigned char sampleRate;    /* Sampling rate                  */
} mouse_state, * pmouse_state;
   

/* forward decl */

static int  mousedev_init(input_module_t *module);
static int  mousedev_reset(input_module_t *module);
static int  mousedev_pulse(message_context_t *, int, unsigned, void *);
static int  mousedev_devctrl(input_module_t *module, int event, void *ptr);
static int  mousedev_parm(input_module_t *module, int opt, char *optarg);
static int  mousedev_shutdown(input_module_t *module, int ms);
static int  read_pnp(struct data *dp, int restoreRate);
static int  get_mouse_type(struct data *dp, int restoreRate);
static int  set_mouse_type(struct data *dp, int type, int restoreRate);
static int  get_mouse_state(struct data *dp, pmouse_state pState);
static int  get_aux_sample_rate(struct data *dp);
static void set_aux_sample_rate(struct data *dp, int nNewRate);
static int  get_aux_resolution(struct data *dp);
static void set_aux_resolution(struct data *dp, int nNewResolution);
static void enable_aux_device(struct data *dp);
static int  reset_aux_device(struct data *dp);


/* PnP request stuff                                */
#define MOUSE_ID_LENGTH          (8) 
/* Mouse responses to PnP ID request with a sequence of 8 scan codes */
/* Following table should be used for transation scan codes to ASCII */
static unsigned char ScanCodeToUChar[] = {
    0x00,            // Nothing
    0x00,            // Esc
    '1',
    '2',
    '3',
    '4',
    '5',
    '6',
    '7',
    '8',
    '9',
    '0',
    '-',
    '=',
    0x00,           // Backspace
    0x00,           // Tab
    'Q',
    'W',
    'E',
    'R',
    'T',
    'Y',
    'U',
    'I',
    'O',
    'P',
    '[',
    ']',
    '\\',
    0x00,            // Caps lock
    'A',
    'S',
    'D',
    'F',
    'G',
    'H',
    'I',
    'J',
    'K',
    'L',
    ';',
    '\'',
    0x00,           // Return
    0x00,           // Shift left
    'Z',
    'X',
    'C',
    'V',
    'B',
    'N',
    'M',
    ',',
    '.',
    '/'
    };

static const int ScanCodeToUCharCount = sizeof(ScanCodeToUChar)/sizeof(unsigned char);

/* Mouse control sequences            */

/* Set wheel + 3 buttons ON           */
static   unsigned char wheel3OnCommands[] =
     {
	 AUX_SAMPLE_RATE, AUX_200_SMP_SEC,
	 AUX_SAMPLE_RATE, AUX_100_SMP_SEC,
	 AUX_SAMPLE_RATE, AUX_80_SMP_SEC
     };
/* Set wheel + 5 buttons ON           */ 
static   unsigned char wheel5OnCommands[] =
     {
	 AUX_SAMPLE_RATE, AUX_200_SMP_SEC,
	 AUX_SAMPLE_RATE, AUX_200_SMP_SEC,
	 AUX_SAMPLE_RATE, AUX_80_SMP_SEC
     };


/* Th following input_module_t data structure represents our device   */
/* module in                                                          */ 

input_module_t	mousedev = {
	NULL,
	NULL,
	NULL,
	0,
	DEVI_CLASS_NONE|DEVI_MODULE_TYPE_DEVICE,
	"mousedev",
	__DATE__,
	"P:rp:i:f:",
	NULL,		/* initialized by us (in init) */
	mousedev_init,
	mousedev_reset,
	NULL,		/* this moduleulator is input & EVBL_TERMINATE */
	NULL,
        mousedev_pulse,
	mousedev_parm,
	mousedev_devctrl,
	mousedev_shutdown,
};









/* Description: this callback funtion is called when the module is linked into the  */
/*              event bus;it is used to set initial module state on the device level*/
/* Input      : input_module_t * module - pointer to module descriptor              */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int mousedev_reset(input_module_t *module)
{
   
   struct data	*dp   = module->data;
   
   // Common reset part
   if((dp->flags & FLAG_INIT) == 0) 
     {

     if(init_contr(dp->flags & FLAG_RESET_KB ? INIT_SELF_TEST : 0,
		   SET_NONE, 
		   dp -> irq,
		   dp -> ioport, dp -> statport))
	 return -1;

     if(register_device(DEFAULT_DISPHANDLE, mousedev_pulse, DEFAULT_COID, AUX_PORT, module))
//     if(register_device(module, AUX_PORT))
	  return -1;
     set_priority(AUX_PORT, dp -> priority);
     dp->flags |= FLAG_INIT;
     }
  
   return reset_aux_device(dp); 

}

/* Description: function resets AUX device                                          */
/* Input      : struct data *dp - pointer to common driver data block               */
/* Output     : None                                                                */
/* Return     : None                                                                */
int reset_aux_device(struct data *dp)
{
        /* Mouse reset command sequence                                             */
   unsigned char commands[] =
     {
     AUX_RESOLUTION, AUX_8_CNTS_MM,
     AUX_ENABLE_XFER
     };
   _uint8 reply[5];
   int i;
   
   reset_device(AUX_PORT);
   //     read_pnp(dp, 0);
   /* get max level of service this device supports                              */
   dp -> type = get_mouse_type(dp, 0);
   if(NOT_DEFINED ==  dp -> type) /* get_mouse_type failed */
       dp -> type = 0; // Set default
   if(NOT_DEFINED ==  dp -> curtype) /* First time accept the max level that device supports */
       dp -> curtype = dp -> type;
   if(dp -> type < dp -> curtype) /* It can be if a mouse was replaced */
       dp -> curtype = dp -> type;
   /* reset mouse */
   /* AUX_RESET is not well documented. Some documents say that there is no ACK for*/
   /* this command (all types of mice I have, reply with ACK). Some documents insist*/
   /* that mouse responses with  AUX_COMPLETE + Mouse type sequence. Besides that, */
   /* I found that if even there was a small delay after AUX_RESET (not more than  */
   /* 10 ms usually),  it also works fine.                                         */
   /* That's why I separated AUX_RESET and inserted reading of up to two bytes to  */
   /* satisfy the case when mouse really wants to say something (of course I just  */
   /* ignore what it says if it says - don't care! If mouse keeps silence, it      */
   /* shouldn't be a problem - we just ignore return code of send_device_command   */
   for(i = 0; i < NUM_OF_RETRIES; ++i)
     {
     send_device_command(AUX_PORT, AUX_RESET, reply, 3);
     switch(reply[0])
	  {
	  case RESEND:
	     continue;
	  default: // ACK or 0xAA?
	     break;
	  }
     break;	
     }
   if(NUM_OF_RETRIES == i) // Cannot reset device
     return -1;
   send_command_set(AUX_PORT, commands, sizeof(commands), NULL, 0);
   // First time if rate and resolution were not changed in the command line
   // keep default values
   if(NOT_DEFINED == dp -> auxRate)
       dp -> auxRate = get_aux_sample_rate(dp); 
   if(NOT_DEFINED == dp -> auxResolution)
       dp -> auxResolution = get_aux_resolution(dp);
   /* If not initial state (as previous reset established) - set last mouse mode */
   if(dp -> curtype > 0)
       set_mouse_type(dp, dp -> curtype, 0);
   /* set device parameters - sample rate and resolution                         */
   set_aux_sample_rate(dp, dp -> auxRate);
   set_aux_resolution(dp, dp -> auxResolution);
   enable_aux_device(dp);
   return EOK;
}
 

/* Description: function initialize mouse device module                             */
/* Input      : input_module_t * module - pointer to module descriptor              */
/* Output     : None                                                                */
/* Return     : 0 if OK, othervise -1                                               */
int mousedev_init(input_module_t *module)
{
    struct data *dp = module->data;

   if(!module->data) 
     {

      if(!(dp = module->data = calloc(sizeof *dp,  1)))
         return (-1);

      dp->irq           = AUX_IRQ;
      dp->ioport        = KBD_DEF_PORT;
      dp->statport      = dp->ioport + STATUS_PORT_ADD;
      dp -> priority    = 0; /* Use default */
       
      dp->curtype       = NOT_DEFINED;
      dp->auxRate       = NOT_DEFINED;
      dp->auxResolution = NOT_DEFINED;
		
      dp->lastByte = 0;
      return 0;
     }
   return (-1);
}


/* Description: function detects PnP type of AUX device                             */
/* Input      : struct data *dp - pointer to common driver data block               */
/*              int restoreRate - 1 - restore sample rate after oeration, 0 - no    */
/* Output     : None                                                                */
/* Return     : None                                                                */
int read_pnp(struct data *dp, int restoreRate)
{          
   char mouseID[MOUSE_ID_LENGTH];
   int nSampleRate = AUX_100_SMP_SEC;
   int rc = 0;
   /* PnP ID request                      */
   static  unsigned char pnpCommands[] =
     {
	 AUX_SAMPLE_RATE, AUX_20_SMP_SEC, 
	 AUX_SAMPLE_RATE, AUX_40_SMP_SEC,
	 AUX_SAMPLE_RATE, AUX_60_SMP_SEC
     };
 
   if(restoreRate && (0 > (nSampleRate = get_aux_sample_rate(dp))))
       return -1;
  
 
   memset(mouseID, 0, sizeof(mouseID));
   if(send_command_set(AUX_PORT, pnpCommands, sizeof(pnpCommands),  
		       mouseID, sizeof(mouseID)))
       rc = -2;

   /* restore sampling rate */
   if(restoreRate)   
      set_aux_sample_rate(dp, nSampleRate);

    
   if(verbosity >= 2)
      printf("Mouse id is %s \n", mouseID);
   return rc;
    
}


/* Description: function enables  output of  AUX device                             */
/* Input      : struct data *dp - pointer to common driver data block               */
/* Output     : None                                                                */
/* Return     : None                                                                */
void enable_aux_device(struct data *dp)
{ 
   unsigned char enableCommands[] =
     {
     AUX_STREAM_MODE,
     AUX_ENABLE_XFER
     };
   send_command_set(AUX_PORT, enableCommands, sizeof(enableCommands),
		    NULL, 0);
} 

/* Description: function takes from AUX device current sampling rate                */
/* Input      : struct data *dp - pointer to common driver data block               */
/* Output     : None                                                                */
/* Return     : Current sampling rate                                               */
int get_aux_sample_rate(struct data *dp)
{
   mouse_state mstate;
   /* save current state to restore sampling rate  after changing mode */
   if(get_mouse_state(dp, &mstate))
     return -1;
   return mstate.sampleRate;
}

/* Description: function sets mouse sample rate                                     */
/* Input      : struct data *dp - pointer to common driver data block               */
/*              int newRate - new sample rate                                       */
/* Output     : None                                                                */
/* Return     : None                                                                */
void set_aux_sample_rate(struct data *dp, int newRate)
{
   unsigned char setRateCmds[] =
     {
     AUX_SAMPLE_RATE, 0
     };

   setRateCmds[1] = newRate;
   send_command_set(AUX_PORT, setRateCmds, sizeof(setRateCmds), NULL, 0);
   
   dp -> auxRate = newRate;
   if(verbosity >= 3)
       printf("New sample rate = %i\n", newRate);
}

/* Description: function gets mouse resolution                                      */
/* Input      : struct data *dp - pointer to common driver data block               */
/* Output     : None                                                                */
/* Return     : Resolution                                                          */
int get_aux_resolution(struct data *dp)
{
   mouse_state mstate;
   /* save current state to restore sampling rate  after changing mode */
   if(get_mouse_state(dp, &mstate))
     return -1;
   return mstate.resolution;
}

/* Description: function sets mouse sample rate                                     */
/* Input      : struct data *dp - pointer to common driver data block               */
/*              int newResolution - new resolution                                  */
/* Output     : None                                                                */
/* Return     : None                                                                */
void set_aux_resolution(struct data *dp, int nNewResolution)
{
   unsigned char setResolutionCmds[] =
     {
     AUX_RESOLUTION, 0
     };
   dp -> auxResolution = nNewResolution;
   setResolutionCmds[1] = nNewResolution;
   send_command_set(AUX_PORT, setResolutionCmds, 
		    sizeof(setResolutionCmds), NULL, 0);
}

/* Description: function determines mouse type                                      */
/* Input      : struct data *dp - pointer to common driver data block               */
/*              int restoreRate - 1 - restore sample rate after oeration, 0 - no    */
/* Output     : None                                                                */
/* Return     : mouse type (0 - no wheel, 0x03 - 3-button + wheel, 0x04 -4-5-button */
/*              + wheel                                                             */
int get_mouse_type(struct data *dp, int restoreRate)
{

   int i, j;
   unsigned char byte;
   int nWheel = NO_WHEEL_MOUSE;
   unsigned char package[20];  
   int nComArraySize;
   int nSampleRate = 0;

   if(restoreRate && (0 > (nSampleRate = get_aux_sample_rate(dp))))
     return -1;

   nComArraySize = sizeof(wheel3OnCommands);
   memcpy(package, wheel3OnCommands, nComArraySize);
   
   for(i = 0; i < 2; ++i)
     {
     package[nComArraySize] = AUX_DEV_ID;
     for(j = 0; j < NUM_OF_RETRIES; ++j)
	{
	byte = 0;
        if(0 == send_command_set(AUX_PORT, package, nComArraySize + 1, &byte, 1))
	   break;
	}
     if(verbosity >= 3)
        printf("Set mouse type as %s; get ID = %#02x\n",
	       (i == 0) ? "3-button" : "5-button", (int)byte);
     if(byte == WHEEL_3B_MOUSE || byte == WHEEL_5B_MOUSE)
         nWheel = byte;
     else 
	 break;
     nComArraySize = sizeof(wheel5OnCommands);
     memcpy(package, wheel5OnCommands, nComArraySize);
     }

   /* restore sampling rate if necessary */
   if(restoreRate)
       set_aux_sample_rate(dp, nSampleRate);
   if(verbosity >= 3)   
       printf("Mouse type = %i\n", nWheel);
   return nWheel;
}

/* Description: function sets mouse type                                            */
/* Input      : struct data *dp - pointer to common driver data block               */
/*              int type - desirable mouse type (see comments to get_mouse_type)    */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int set_mouse_type(struct data *dp, int type, int restoreRate)
{
   int rc = 0;
   int nSampleRate = 0;

   if(restoreRate && (0 > (nSampleRate = get_aux_sample_rate(dp))))
     return -1;
   if((0 > type) || (type > dp->type)) /* This mouse doesn't support this mode */
     return -1;

   dp->curtype = type;
   switch(type)
     {
      case NO_WHEEL_MOUSE:    /* mouse without wheel   */
	reset_aux_device(dp);
	break;
      case WHEEL_3B_MOUSE: /* 3-buttons wheel mouse */
        send_command_set(AUX_PORT, wheel3OnCommands, 
			 sizeof(wheel3OnCommands), NULL, 0);
	break;
      case WHEEL_5B_MOUSE: /* 4-5-buttons wheel mouse */
	/* do this because some switch boxes don't permit to jump from 0-mode to    */
	/* 4-mode                                                                   */
        send_command_set(AUX_PORT, wheel3OnCommands, 
			 sizeof(wheel3OnCommands), NULL, 0);
        send_command_set(AUX_PORT, wheel5OnCommands, 
			 sizeof(wheel5OnCommands), NULL, 0);
	break;
      default:
	rc = -1;
        break; 
     }
   
   /* restore sampling rate */
   if(restoreRate)
       set_aux_sample_rate(dp, nSampleRate);
   return rc;
}

/* Description: function gets mouse state                                           */
/* Input      : struct data *dp - pointer to common driver data block               */
/* Output     : pmouse_state pState - pointer to structure to keep mouse state      */
/* Return     : 0 if OK, otherwise -1                                               */
int get_mouse_state(struct data *dp, pmouse_state pState)
{
   unsigned char * buf;
   
   buf = (unsigned char *)pState;

   memset(buf, 0, sizeof(mouse_state));
   
   return send_command(AUX_PORT, AUX_GET_STATUS, buf, 3);
   
}


/* Description: this is a callback that PS2 engine activates asychroniously each    */
/*              time as keyboard or mouse interrupt occures                         */
/* Input      : message_context_t *ctp - pointer to module descriptor               */
/*              int code - program doesn't use this parameter                       */
/*              unsigned flags - program doesn't use this parameter                 */
/*              void *data - pointer to (input_module_t	mouse) structure            */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int mousedev_pulse(message_context_t *ctp, int code, unsigned flags, void *data)
{
    input_module_t		*module = (input_module_t *) data;
    struct data 		*dp = module->data;
    input_module_t 		*up = module->up;
    unsigned char 		c = get_in_data(ctp);

    if ((AUX_COMPLETE == dp->lastByte) && 
        ((NO_WHEEL_MOUSE == c) || 
         (WHEEL_3B_MOUSE == c) || 
         (WHEEL_5B_MOUSE == c)))
	{
        if(verbosity >= 4)
	   fprintf(stderr, "Mouse was re-plugged. Reset it!\n");
	dp -> curtype = -1; /* Specially to take now the highest service level */   
	if(up->reset)
	   (up->reset)(up);
	else   
	   reset_aux_device(dp);
	return 0;
	}
   
     if (AUX_COMPLETE == c) /* May be re-plug; wait next byte */
        {
        dp->lastByte = c;
	return 0;
        }
   

     dp->lastByte = c;
  
    
     if (verbosity >= 5) 
        printf("\tSending to protocol: %#02x\n", (int)c);
    
     (up->input)(up, 1, (void *) &c);
     if(0 < dp -> fId)
	write(dp -> fId, &c, 1);
  
        
    return (0);
} 
    

/* Description: this is a callback function for DEVCTRL command processing          */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int event  - DEVCTRL command code                                   */
/*              void * ptr - pointer to data exchange block                         */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int mousedev_devctrl(input_module_t *module, int event, void *ptr)
{
        struct data *dp = module->data;

        switch(event) {


        case DEVCTL_GETMOUSETYPE:
	   /* Get mouse type               */
	     if(NULL != ptr)
	        {
                struct devctl_mouse_types * pmt;	    
	        pmt = (struct devctl_mouse_types *)ptr;
                pmt -> curtype = dp -> curtype;
		pmt -> type    = dp -> type;
   	        }
	     break;
	 case DEVCTL_SETMOUSETYPE:
	   /* Set mouse type                */
	     if(NULL != ptr)
                return set_mouse_type(dp, *(int *)ptr, 1);		  
	case DEVCTL_GETSAMPLE_RATE:   
	   if(NULL != ptr)
	     *(int *)ptr = dp -> auxRate;
	   break;
	case DEVCTL_SETSAMPLE_RATE:
	   if(NULL != ptr)
	     {
             int nNewSampleRate = *(int *)ptr;
	     if(nNewSampleRate > 0)
		{
	        dp -> auxRate = nNewSampleRate;
	        if(module -> up -> reset)
	           (module->up->reset)(module->up);
	        else
	           reset_aux_device(dp);
		}
	     }
	   break;
	
	case DEVCTL_GET_RESOLUTION:
	   if(NULL != ptr)
	     *(int *)ptr = dp -> auxResolution;
	   break;
	case DEVCTL_SET_RESOLUTION:
	   if(NULL != ptr)
	     {
	     int nNewResolution = *(int *)ptr;
	     switch(nNewResolution)
		  {
		   case 1:
		     nNewResolution = AUX_1_CNTS_MM;
		     break;
		   case 2:
		     nNewResolution = AUX_2_CNTS_MM;
		     break;
		   case 4:
		     nNewResolution = AUX_4_CNTS_MM;
		     break;
		   case 8:
		     nNewResolution = AUX_8_CNTS_MM;
		     break;
		   default:
		     return (-1);
		  }
	     dp -> auxResolution = nNewResolution;
	     if(module -> up -> reset)
	        (module->up->reset)(module->up);
	     else
	        reset_aux_device(dp);	 
	     }
	   break;
        default:
                return (-1);
        }

        return (EOK);
}


/* Description: this is a callback function for command line parameter processing   */
/*              (all valid parameters for device module are listed in mouse.args)   */ 
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int opt  - parameter cde                                            */
/*              char * optarg - optional parameter value                            */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int mousedev_parm(input_module_t *module, int opt, char *optarg)
{
        struct data *dp = module->data;


        switch (opt) {

        case 'r':
                dp->flags |= FLAG_RESET_KB;
                break;
                
        case 'i':
                dp->irq = strtoul(optarg, 0, 0);
                break;
                
        case 'p':
	        if(NULL != optarg)
	           {
		   char * p = strchr(optarg, ',');
                   dp->ioport = strtoul(optarg, 0, 16);
		   if(NULL != p)
		      dp->statport = dp->ioport  + strtol(p + 1, 0, 10);
		   else
		      dp->statport = dp->ioport + STATUS_PORT_ADD;
		   }
                break;
	case 'P':
	       if(optarg)
	          {
		  int nPrty = atoi(optarg);
		  if(nPrty > 0) 
	             dp -> priority = nPrty;
		  }
                break;
     	  case 'f':
	        dp -> fId = open(optarg, O_CREAT | O_WRONLY | O_TRUNC,
				 S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
	        break;
        default:
                return (-1);
        }
        
        return (0);
}
 
/* Description: this is a callback function which is called when resourse manager   */
/*              is shutting down                                                    */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int ms  - program doesn't use this parameter                        */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int mousedev_shutdown(input_module_t *module, int ms)
  {
    struct data *pd = module->data;
 
    unregister_device(AUX_PORT);
    if(0 < pd -> fId) /* if debug output file was created - close it */
       close(pd -> fId);
    return (0);
  }





 
