/*
 * Copyright  2000, QNX Software Systems Ltd.  All Rights Reserved
 *
 * This source code has been published by QNX Software Systems Ltd.
 * (QSSL).  However, any use, reproduction, modification, distribution
 * or transfer of this software, or any software which includes or is
 * based upon any of this code, is only permitted under the terms of
 * the QNX Open Community License version 1.0 (see licensing.qnx.com
 * for details) or as otherwise expressly authorized by a written license
 * agreement from QSSL.  For more information, please email licensing@qnx.com.
 */


/*
 * msoft.c
 *
 * The Microsoft mouse protocol.  Covers standard 2 button as
 * well as 3 button and "Intelli" wheel mice. This module is applicable
 * for seral mice only
 *
 */
#include "devi.h"

#define FLAG_INIT		0x0100
#define FLAG_3BUTTON		0x0200
#define FLAG_MIDDLE		0x0400
#define FLAG_RESET		0x0800
#define FLAG_INTELLIMSE		0x1000
#define FLAG_WHEEL_ON           0x2000

static int msoft_init(input_module_t *module);
static int msoft_devctrl(input_module_t *module, int event, void *ptr);
static int msoft_reset(input_module_t *module);
static int msoft_input(input_module_t *module, int num, void *arg);
static int msoft_parm(input_module_t *module,int opt,char *optarg);
static int msoft_shutdown(input_module_t *module, int delay);

/* Our protocol module is represented by the following input_module_t  data structure */

input_module_t	msoft =
{
        NULL,
	NULL,
	NULL,
	0,
	DEVI_CLASS_REL|DEVI_MODULE_TYPE_PROTO,
	"MSoft",
	__DATE__,
	"iR3b:",
	NULL,
	msoft_init,
	msoft_reset,
	msoft_input,
	NULL,
	NULL,
	msoft_parm,
	msoft_devctrl,
	msoft_shutdown
};

/* Protocol module private data                                                                      */
struct private_data
{
  int			flags;
  struct packet_rel	mp;              /* data buffer                                              */
  int			baud;            /* serial port baud rate                                    */
  int			state;           /* current state (see msoft_input for details)              */
  int			packet_size;
};

/* Description: callback initialisation function; it is called when input module    */
/*              initializes input system                                            */
/* Input      : input_module_t * module - pointer to module descriptor              */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise - (-1)                                           */
int   msoft_init(input_module_t *module)
{
  struct private_data *dp = module->data;

  if(!module->data)
	{
	  if(!(dp = module->data = scalloc(sizeof *dp)))
		{
		  return (-1);
		}
	  dp->baud = 1200;

	/* default is to reset mouse */
	  dp->flags = FLAG_RESET;
	  dp->packet_size = 3;
	}
  return (0);
}

/* Description: this is a callback function for DEVCTRL command processing          */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int event  - DEVCTRL command code                                   */
/*              void * ptr - pointer to data exchange block                         */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int  msoft_devctrl(input_module_t *module, int event, void *ptr)
{
  struct private_data 	*dp = module->data;
  input_module_t		*down = module->down;

  switch(event)
	{

	  case	DEVCTL_GETDEVFLAGS:
	     *(unsigned short *)ptr=(dp->flags & FLAGS_GLOBAL);
	     break;

	  case	DEVCTL_GETPTRBTNS:
	     if(dp->flags & FLAG_3BUTTON)
		{
		  *(unsigned long *)ptr=7L;
		}
	     else
		{
		  *(unsigned long *)ptr=5L;
		}
	     break;

	  case	DEVCTL_GETPTRCOORD:
	     *(unsigned char *)ptr='\02';
	     break;
	  case DEVCTL_GETWHEEL:
	     if((dp->flags & FLAG_INTELLIMSE) && (NULL != ptr))
	        {
		if(dp->flags & FLAG_WHEEL_ON)
		   *(int *)ptr = 3;
		else
		   *(int *)ptr = 2;
	        }
	     else
	        return -1;
	     break; 
	  case DEVCTL_SETWHEEL:
	     if((dp->flags & FLAG_INTELLIMSE) && (NULL != ptr))
	        {
		if(*(int *)ptr & 1) /* ON */
		   dp->flags |= FLAG_WHEEL_ON;
		else
		   dp->flags &= ~FLAG_WHEEL_ON;
	        }
	     else
	        return -1;
	     break;
	  default:
	     if(down && down->devctrl)
		{
		  return((down->devctrl)(down, event, ptr));
		}
	     return (-1);
	}

  return (0);
}

/* Description: this callback funtion is called when the module is linked into the  */
/*              event bus;it is used to set initial module state on the protocol    */
/*              level                                                               */
/* Input      : input_module_t * module - pointer to module descriptor              */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int  msoft_reset(input_module_t *module)
{
  struct devctl_setbaud	devctl;
  struct devctl_settrig	trig;
  struct private_data	*dp = module->data;
  input_module_t		*down = module->down;

  if((dp->flags & FLAG_INIT) == 0)
	{

	  devctl.baud = dp->baud;
	  devctl.cflag = PNONE | STOP2 | CS7;

	  if(down && down->devctrl)
		{

		  (down->devctrl)(down, DEVCTL_SETBAUD, &devctl);

		  trig.trigger_level=dp->packet_size;
		  (down->devctrl)(down, DEVCTL_SETTRIG, &trig);

		  if(dp->flags & FLAG_RESET)
			{
			  (down->devctrl)(down, DEVCTL_RESET, NULL);
			}

		}
	  dp->flags |= FLAG_INIT;
	}
  return (0);
}

/* Description: main protocol processing function.  It will be called  by the       */
/*              device layer to pass it data to process.  Its job is to interpret   */
/*              the data according to the MS wheel/not wheel mouse protocol, create */
/*              a data structure, fill it in and hand it off to the  filter layer   */
/*              module above it. The protocol is processed using simple state       */
/*              machine.  State info is kept in module's private data               */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int num - number of bytes to process                                */
/*              void * arg - raw data to process                                    */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int msoft_input(input_module_t *module, int num, void *arg)
{
  int			        i, enqueue = 0;
  struct private_data		*dp = module->data;
  input_module_t			*down = module->down;
  input_module_t			*up = module->up;
  struct devctl_settrig		trig;
  char 				*p = (char *) arg;

  if(dp->flags & FLAG_INIT)
	{

	  for(i = 0; i < num; i++, p++)
		{
		if(!(dp->flags & FLAG_INTELLIMSE))
		     {
                       
    		     if(*p & 0x40)
			  {
			  if(!(dp->flags & FLAG_3BUTTON))
			      {
			      dp->flags &= ~FLAG_MIDDLE;
			      }
			  dp->state = 0;

			  }
		     else if (dp->state == 0)
			  {
			  if((*p & 0x5C) == 0)
			      {
			      if(*p & 0x20)
				   {
				   dp->mp.buttons |= _POINTER_BUTTON_MIDDLE;
				   dp->flags |= FLAG_MIDDLE;
				   }
			      else
				   {
				   dp->flags &= ~FLAG_MIDDLE;
				   }
			      enqueue = 1;

			      }
			 else
			      {
			      continue;			/* framing error */
			      }
			}
		 }
	     else 
	     if (*p & 0x40) /* sync up the intellimouse */
		 dp->state = 0;

	      switch(dp->state)
		 {

		 case 0:
		     if(enqueue)
			 break;
		     dp->mp.buttons = ((*p & 0x20) ? _POINTER_BUTTON_LEFT : 0) |
				((*p & 0x10) ? _POINTER_BUTTON_RIGHT : 0);
		     dp->mp.dy = ((*p & 0x08) ? -1 : 0);
		     dp->mp.dx = ((*p & 0x02) ? -1 : 0);
		     dp->mp.dy |= (*p & 0x0c) << 4;
		     dp->mp.dx |= (*p & 0x03) << 6;
		     dp->state++;
		     break;

		 case 1:
		     dp->mp.dx = (dp->mp.dx & ~0x003f) |
					(*p & 0x003f);
		     dp->state++;
		     break;

		case 2:
		     dp->mp.dy = (dp->mp.dy & ~0x003f) |
					(*p & 0x003f);
		     if(dp->flags & FLAG_INTELLIMSE)
		      	{
			dp->mp.flags |= PTR_Z_DATA;
			dp->state = 3;
			}
		     else
			{
			dp->mp.flags &= ~PTR_Z_DATA;
			if(!(dp->flags & FLAG_MIDDLE))
			   {
			   enqueue = 1;
			   }
		        dp->state=0;
			}
		    break;

		 case 3:
		    dp->mp.dz = (*p & 0x01);
		    if(*p & 0x08)
			{
			dp->mp.dz = -dp->mp.dz;
			}
		    if(*p & 0xf0)
			{
			dp->mp.buttons |= _POINTER_BUTTON_MIDDLE;
			}
		    else
			{
			dp->mp.buttons &= ~_POINTER_BUTTON_MIDDLE;
			}
		    enqueue = 1;
		    break;
		 }

              /* pass data to filter module                                   */ 
	      if(enqueue)
		 {
		 dp->mp.dy = -(dp->mp.dy);
		 clk_get(&dp->mp.timestamp);
		 (up->input)(up, 1, &dp->mp);
		 enqueue = 0;
		 dp->state = 0;
		 }
	  }

	  if(dp->flags & FLAG_3BUTTON)
		{
		  trig.trigger_level = 1;
		  (down->devctrl)(down, DEVCTL_SETTRIG, &trig);
		}
	  else
		{

		  trig.trigger_level = (dp->flags & FLAG_MIDDLE) ? 4 :
		  dp->packet_size - dp->state;

		  (down->devctrl)(down, DEVCTL_SETTRIG, &trig);
		}
	}

  return (0);
}

/* Description: this is a callback function for command line parameter processing   */
/*              (all valid parameters for device module are listed in msoft.args)   */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int opt  - parameter code                                           */
/*              char * optarg - optional parameter value                            */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
/* Comment    : unlike to PS2 mouse, serial mouse doesn't supply any information    */
/*              about its features, that's why command line parameters should be    */
/*              used to set serial mouse to specific mode                           */
int msoft_parm(input_module_t *module,int opt,char *optarg)
{
  struct private_data *dp = module->data;

  switch (opt)
	{

	  case 'i':
	  dp->packet_size = 4;
	  dp->flags |= FLAG_INTELLIMSE | FLAG_WHEEL_ON;
	  break;

	  case 'R':
	  dp->flags &= ~FLAG_RESET;
	  break;

	  case '3':
	  dp->flags |= FLAG_3BUTTON;
	  break;

	  case 'b':
	  dp->baud = strtol(optarg,0,0);
	  break;

	  default:
	  return (-1);
	}

  return (0);
}

/* Description: this is a callback function which is called when resourse manager   */
/*              is shutting down                                                    */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int ms  - program doesn't use this parameter                        */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
/* Comment    : Does nothing for the protocol level                                 */
int  msoft_shutdown(input_module_t *module, int delay)
{
  return (0);
}
