/*
 * Copyright  2000, QNX Software Systems Ltd.  All Rights Reserved
 *
 * This source code has been published by QNX Software Systems Ltd.
 * (QSSL).  However, any use, reproduction, modification, distribution
 * or transfer of this software, or any software which includes or is
 * based upon any of this code, is only permitted under the terms of
 * the QNX Open Community License version 1.0 (see licensing.qnx.com
 * for details) or as otherwise expressly authorized by a written license
 * agreement from QSSL.  For more information, please email licensing@qnx.com.
 */

/* 
 * msys.c
 *
 * The Mouse systems mouse protocol.
 *
 */
#include "devi.h"

#define FLAG_INIT			0x0100
#define FLAG_VALID			0x0200
#define FLAG_RESET			0x0400

/* Forward declarations                                                 */
static int msys_init(input_module_t *module);
static int msys_devctrl(input_module_t *module, int event, void *ptr);
static int msys_reset(input_module_t *module);
static int msys_input(input_module_t *module, int num, void *arg);
static int msys_parm(input_module_t *module,int opt,char *optarg);
static int msys_shutdown(input_module_t *module, int delay);

/* Our protocol module is represented by the following input_module_t  data structure */

input_module_t	msys = {
	NULL,
	NULL,
	NULL,
	0,
	DEVI_CLASS_REL|DEVI_MODULE_TYPE_PROTO,
	"MSys",
	__DATE__,
	"Rb:",
	NULL,			// initialized by us (in init)
	msys_init,
	msys_reset,
	msys_input,
	NULL,
	NULL,
	msys_parm,
	msys_devctrl,
	msys_shutdown
};

/* Protocol module private data                                                                      */

struct data {
	unsigned 		flags;
	struct packet_rel	mp;       /* data buffer                                              */
	int			state;    /* current state (see msoft_input for details)              */
        int			baud;     /* serial port baud rate                                    */
};


/* Description: callback initialisation function; it is called when input module    */
/*              initializes input system                                            */
/* Input      : input_module_t * module - pointer to module descriptor              */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise - (-1)                                           */
int msys_init(input_module_t	*module)
{
	struct data *cptr=module->data;

	if(!module->data)  {
		if(!(cptr = module->data = scalloc(sizeof *cptr))) {
			return -1;
		}
		cptr->flags = FLAG_RESET;
		cptr->baud = 1200;
	}

	return(0);
}

/* Description: this is a callback function for DEVCTRL command processing          */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int event  - DEVCTRL command code                                   */
/*              void * ptr - pointer to data exchange block                         */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int msys_devctrl(input_module_t *module,int event,void *ptr)
{
	input_module_t	*down=module->down;

	if(down && down->devctrl) {
		return((down->devctrl)(down, event, ptr));
	} else {
		return -1;
	}
}

/* Description: this callback funtion is called when the module is linked into the  */
/*              event bus;it is used to set initial module state on the protocol    */
/*              level                                                               */
/* Input      : input_module_t * module - pointer to module descriptor              */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int msys_reset(input_module_t *module)
{
	struct devctl_setbaud	portctl;
	struct devctl_settrig	trig;
	struct data				*cptr=module->data;
	input_module_t			*down=module->down;

	if((cptr->flags & FLAG_INIT) == 0) {

		if(down && down->devctrl) {
			portctl.baud = cptr->baud;
			portctl.cflag = PNONE | STOP1 | CS8;
			(down->devctrl)(down, DEVCTL_SETBAUD, &portctl);
			
			if(cptr->flags & FLAG_RESET) {
				(down->devctrl)(down, DEVCTL_RESET, 0);
			}

			trig.trigger_level=5;  // five byte protocol
			(down->devctrl)(down, DEVCTL_SETTRIG, &trig);

		}

		cptr->state = 0;
		cptr->flags |= FLAG_INIT;
	}

	return(0);
}

/* Description: main protocol processing function.  It will be called  by the       */
/*              device layer to pass it data to process.  Its job is to interpret   */
/*              the data according to the Mouse System  mouse protocol, create      */
/*              a data structure, fill it in and hand it off to the  filter layer   */
/*              module above it. The protocol is processed using simple state       */
/*              machine.  State info is kept in module's private data               */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int num - number of bytes to process                                */
/*              void * arg - raw data to process                                    */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
/* Comment    : driver uses one of two input modules depending on device data       */
/*              protocol                                                            */  
int msys_input(input_module_t *module, int num, void *ptr)
{
	int				i;
	char				*p=(char *)ptr;
	struct data			*cptr=module->data;
	input_module_t			*down=module->down;
	input_module_t			*up=module->up;
	struct devctl_settrig		trig;

	if(cptr->flags & FLAG_INIT) {
		for(i = 0; i < num; i++,p++) {
			if(cptr->state == 0 && (*p & 0xf8) != 0x80) {
				continue;
			} 
			switch(cptr->state) {
				case 0:
					cptr->mp.buttons = (~(*p) & 0x07);
					cptr->state++;
					break;
				case 1:
					cptr->mp.dx = (signed char)*p;
					cptr->state++;
					break;
				case 2:
					cptr->mp.dy = (signed char)*p;
					cptr->state++;
					break;
				case 3:
					cptr->mp.dx += (signed char)*p;
					cptr->state++;
					break;
				case 4:
					cptr->mp.dy += (signed char)*p;
					clk_get(&cptr->mp.timestamp);
                                /* pass data to filter module        */ 
					(up->input)(up, 1, &cptr->mp);
			        /* and reset state machine           */
					cptr->state = 0;
					break;
				default:
					break;
			}

		}

		if(down && down->devctrl) {
			trig.trigger_level = cptr->state ? 5 - cptr->state : 1;
			(down->devctrl)(down, DEVCTL_SETTRIG, &trig);
		} 
	}

	return(0);
}


/* Description: this is a callback function for command line parameter processing   */
/*              (all valid parameters for device module are listed in msys.args)    */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int opt  - parameter code                                           */
/*              char * optarg - optional parameter value                            */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int msys_parm(input_module_t *module,int opt,char *optarg)
{
	struct data *cptr=module->data;

 	switch (opt) {
		case 'R':
			cptr->flags &= ~FLAG_RESET;
 			break;
		case 'b':
			cptr->baud=strtol(optarg,0,0);
 			break;
 		default:
 			return -1;
 	}

	return(0);
}


/* Description: this is a callback function which is called when resourse manager   */
/*              is shutting down                                                    */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int ms  - program doesn't use this parameter                        */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
/* Comment    : Does nothing for the protocol level                                 */
int msys_shutdown(input_module_t *module, int delay)
{
	return(0);
}

