
/* Copyright, 1999-2000, QNX Software Systems Ltd.  All Rights Reserved

 * This source code has been published by QNX Software Systems
 * Ltd. (QSSL). However, any use, reproduction, modification, distribution
 * or transfer of this software, or any software which includes or is based
 * upon any of this code, is only permitted if expressly authorized by a
 * written license agreement from QSSL. Contact the QNX Developer's Network
 * or contact QSSL's legal department for more information.
 */


/*
 * penmount.c
 *
 *  PenMount touch screen  protocol. 
 *  9584A controller card, chip 9512-B5.5 (Salt International Corp.)
 *  touch screen:Dynapro 95402 (3M Dynapro)
 *   A.Smolenkov
 */

#include "devi.h"

#define	TRUE					1
#define	FALSE					!TRUE
#define FLAG_INIT				0x1000
#define FLAG_RESET				0x2000
#define DEBOUNCE			 	80000L


struct data {
	int					flags;
        struct packet_abs	                tp;       
	int					baud;     /* data transfer baud rate */
	int					state;    /* current state (see smart_input for details) */
	int					pktsize;
	int                 init_delay;  /* art */
};

static int pen_init(input_module_t *module);
static int pen_devctrl(input_module_t *module,int event,void *ptr);
static int pen_reset(input_module_t *module);
static int pen_input(input_module_t *module, int num, void *ptr);
static int pen_parm(input_module_t *module,int opt,char *optarg);
static int pen_shutdown(input_module_t *module, int delay);


input_module_t	penmount = {

	NULL,
	NULL,
	NULL,
	0,
	DEVI_CLASS_ABS|DEVI_MODULE_TYPE_PROTO,
	"penmount",
	__DATE__,
	"Rb:d:",
	NULL,			// initialized by us (in init)
	pen_init,
	pen_reset,
	pen_input,
	NULL,
	NULL,
	pen_parm,
	pen_devctrl,
	pen_shutdown,
};


/* Description: callback initialisation function; it is called when input module    */
/*              initializes input system                                            */
/* Input      : input_module_t * module - pointer to module descriptor              */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise - (-1)                                           */
int pen_init(input_module_t *module)
{
        struct data *dp = module->data;
  
        if(!module->data)  {
                
                if(!(dp = module->data = calloc(sizeof *dp, 1))) {
                        return (-1);
                }
                dp->flags = ABSOLUTE | ABS_UNCALIBRATED | FLAG_RESET;
                dp->baud = 19200;//!! 
                dp->state = 0;
                dp->init_delay = 100; // art
        }
        
        return (0);
}


/* Description: this is a callback function for DEVCTRL command processing          */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int event  - DEVCTRL command code                                   */
/*              void * ptr - pointer to data exchange block                         */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int pen_devctrl(input_module_t *module, int event, void *ptr)
{
	struct data 	*dp = module->data;
	input_module_t	*down = module->down;

	switch(event) {

		case DEVCTL_GETDEVFLAGS:
			*(unsigned short *)ptr = (dp->flags & FLAGS_GLOBAL);
			break;
		case DEVCTL_GETPTRBTNS:
			*(unsigned long *)ptr = 1L;
			break;
		case DEVCTL_GETPTRCOORD:
			*(unsigned char *)ptr = '\02';
			break;
	 	case DEVCTL_GETPTRPRESS:
			*(unsigned char *)ptr = (dp->flags & ABS_PRESS_DATA)?1:0;
			break;
	 	case DEVCTL_GETCOORDRNG: {
			struct devctl_coord_range	*range = ptr;

			// we don't know this, pass it on maybe the device knows ?
			if((down->devctrl)(down, DEVCTL_GETCOORDRNG, ptr) == -1) {
				// the lower layer doesn't know.  probably a serial device
				// well, let's try and guess.  If there is pressure data
				// then it must be 12 bit.  Otherwise we'll assume that
				// it's 8 bit
				if(dp->flags & ABS_PRESS_DATA) {
					range->min = 0;
					range->max = 4095;
				} else {
					range->min = 0;
					range->max = 255;
				} 
			} 
			break;
		}
	 	case DEVCTL_GETPRESSRNG: {
			// for all smartset devices at present the range is 0-255
			struct	devctl_coord_range	*range = ptr;

			range->min = 0;
			range->max = 255;
			break;
		}
		default:
			if(down && down->devctrl) {
				return((down->devctrl)(down, event, ptr));
			} else {
				return (-1);
			}
	}

	return (0);
}

        
/* Descr: Init PenMount 9584A controller                   */
/*   Inp:   down - ptr to device module, del_val - delay   */
int send_init_seq(input_module_t *down,int del_val)
 {
   char *header="\x44\x47";  //"DG"
  // PenMount init codes seq (without headers (0x44,0x47))
  // command=header,code1,code2,code3 
   static unsigned char init_seq[] = 
       {0x0a,0x0,0x0, 0x2,0x3,0xfe, 0x2,0x13,0xfe, 0x03,0x09,0x0,
        0x4,0xff,0x0, 0x0,0xf7,0x0f,0x0a,0x1,0x0 };
   int cntr=0;
   char *next;
   
   next=init_seq;
   for(cntr=0;cntr< (sizeof init_seq / 3);cntr++,next+=3)
    {    
     (down->output)(down, header,  2);
     (down->output)(down, next, 3);   
     delay(del_val);
    }
   return 0;
 }

/* Description: this callback funtion is called when the module is linked into the  */
/*              event bus;it is used to set initial module state on the protocol    */
/*              level                                                               */
/* Input      : input_module_t * module - pointer to module descriptor              */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int pen_reset(input_module_t *module)
{
        struct data			*dp = module->data;
        input_module_t			*down = module->down;
        struct devctl_setbaud	        portctl;
        struct devctl_settrig	        trig;
        
        if((dp->flags & FLAG_INIT) == 0) {
                
                if(down && down->devctrl) {
                        
                        portctl.baud = dp->baud;
                        portctl.cflag = PNONE | STOP2 | CS8;
                        (down->devctrl)(down, DEVCTL_SETBAUD, &portctl);
                        
                        if(dp->flags & FLAG_RESET) {
                            (down->devctrl)(down, DEVCTL_RESET, NULL);
                            if (verbosity)
                               printf("penmount: reset fd\n");    
                        }
                        
                        trig.trigger_level = dp->pktsize = 1; //!
                        (down->devctrl)(down, DEVCTL_SETTRIG, &trig);
                        
                }
                
                dp->state = 0;
                dp->flags |= FLAG_INIT;
        }
        
        if(down && down->output) {
                      if (verbosity)
                                printf("penmount: send_init_seq (delay=%d)\n",dp->init_delay);     
                      send_init_seq(down,dp->init_delay);
           }
  /*    //???  
        if(dp->flags & FLAG_RESET) {
                            (down->devctrl)(down, DEVCTL_RESET, NULL);
                            if (verbosity)
                                printf("penmount: reset fd\n");    
        }  */
        
        return (0);
}


/* Description: main protocol processing function.  It will be called  by the       */
/*              device layer to pass it data to process.  Its job is to interpret   */
/*              the data according to the PenMount touch screen protocol, create    */
/*              a data structure, fill it in and hand it off to the  filter layer   */
/*              module above it. The protocol is processed using simple state       */
/*              machine.  State info is kept in module's private data               */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int num - number of bytes to process                                */
/*              void * arg - raw data to process                                    */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int pen_input(input_module_t *module, int num, void *ptr)
{
        int					i, enqueue = 0;
        unsigned char			*p = (unsigned char *)ptr;
        struct data				*dp = module->data;
        input_module_t				*up = module->up;
        
        if(dp->flags & FLAG_INIT) {
                
                for (i = 0; i < num; i++, p++) {
                        
                     switch (dp->state) {   
                        case 0: // header (press)
                                if (*p == 0xff)
                                        dp->state++;
                                break;
                        case 1: 
                               if (*p == 0xfd){
                                        dp->state++;
                                    }else{
                                        dp->state=0;
                                }        
                                break;       
                        case 2:
                              if (*p == 0xfd){
                                      dp->state++;
                                  }else{
                                      dp->state=0;
                                }   
                                break; 
                        case 3:   // data
                              if (*p == 0xff){
                                     dp->state++;
                                  }else{
                                     dp->state=0;
                                }
                                break;          
                        case 4: 
                        case 5:
                        case 6:
                        case 7:
                                dp->state++;
                                break;
                        case 8:
                            if (*p == 0xff){
                                     dp->state++;
                                  }else{
                                     dp->state=0;
                                }
                                break;
                        case 9:               
                               if (*p == 0xfe){ // tail (release)
                                  dp->state=13;
                                 }else{
                                  dp->tp.x = ((unsigned char ) *p << 8) & 0xff00;
                                  dp->state++;
                                 }
                                break;        
                        case 10:
                                dp->tp.x |= (unsigned char ) *p;
                                dp->state++;
                                break;
                        case 11:
                                dp->tp.y = ((unsigned char ) *p << 8) & 0xff00;
                                dp->state++;
                                break;        
                        case 12:
                                dp->tp.y |= (unsigned char ) *p;
                                dp->tp.buttons = 4L;  //button pressed
                                dp->tp.z = 0;
                                dp->state = 8;
                                enqueue = 1;
                                break;     
                         case 13:
                               dp->tp.buttons = 0; // button released
                               dp->state = 0;
                               enqueue = 1;
                               break;               
                           }     
             /*  !!!  */                                                                   
                        if (enqueue) {
                                
                                dp->tp.flags = dp->flags;
                                clk_get(&dp->tp.timestamp);
                                enqueue = 0;
                                
                                (up->input)(up, 1, &dp->tp);
                        }
                }
        }
        return (0);
}


/* Description: this is a callback function for command line parameter processing   */
/*                                                                                  */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int opt  - parameter code                                           */
/*              char * optarg - optional parameter value                            */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
int pen_parm(input_module_t *module, int opt, char *optarg)
{
	struct data *dp = module->data;

 	switch (opt) {
		case 'R':
			dp->flags &= ~FLAG_RESET;
			break;
		case 'b':
			dp->baud = atoi(optarg);
			break;
		case 'd':
		    dp->init_delay = atoi(optarg);	
		      if (verbosity){
		         printf("penmount: set delay %s ms\n",optarg);
		        }
		    break;
 		default:
 			return (-1);
 	}

	return (0);
}


/* Description: this is a callback function which is called when resourse manager   */
/*              is shutting down                                                    */
/* Input      : input_module_t * module - pointer to module descriptor              */
/*              int delay  - program doesn't use this parameter                     */
/* Output     : None                                                                */
/* Return     : 0 if OK, otherwise -1                                               */
/* Comment    : Does nothing for the protocol level                                 */
int pen_shutdown(input_module_t *module, int delay)
{
	return (0);
}

