/*
 * Copyright 2001, QNX Software Systems Ltd. All Rights Reserved
 *
 * This source code has been published by QNX Software Systems Ltd. (QSSL).
 * However, any use, reproduction, modification, distribution or transfer of
 * this software, or any software which includes or is based upon any of this
 * code, is only permitted under the terms of the QNX Open Community License
 * version 1.0 (see licensing.qnx.com for details) or as otherwise expressly
 * authorized by a written license agreement from QSSL. For more information,
 * please email licensing@qnx.com.
 */

#include "devi.h"

/* samp_dev.c
 * 
 * A sample device module.
 *
 */


/* module specific flags and constants */

#define FLAG_INIT                0x00001  /* module has been initialized */

#define DEFAULT_IRQ		1
#define DEFAULT_IOPORT		0xff


/* Our modules private data, put stuff like irq, ioport base, flags
 * etc in here */
struct private_data {

        int 			irq; /* irq to attach to */
        int			iid; /* interrupt id */
        long         		ioport;	/* phys address of device */
        uintptr_t             	iobase;	/* mmaped device memory base  */
        unsigned 		flags;
        int			irq_pc; /* irq pulse code */
	int			tim_pc; /* timer pulse code */
        timer_t			timer_id;
	struct itimerspec	ts;

        /* ... also put things like data buffers, etc here */
};


/* forward decl */
static int samp_dev_init(input_module_t *module);
static int samp_dev_reset(input_module_t *module);
static int samp_dev_pulse(message_context_t *, int, unsigned, void *);
static int samp_dev_devctrl(input_module_t *module, int event, void *ptr);
static int samp_dev_parm(input_module_t *module, int opt, char *optarg);
static int samp_dev_shutdown(input_module_t *module, int ms);
static int samp_dev_output(input_module_t *module, void *data, int count);

/* Our device module is represented by the following input_module_t
 * data structure.  We always create one static instance of the module.
 * If more than one are needed, ie in multiple bus lines; then the system
 * will allocate a new module and copy the contents of the static one
 * into it.
 */

input_module_t	samp_dev = {
        NULL,		        /* up module in bus, filled in at runtime */
	NULL,			/* down module in busline */
	NULL,			/* event bus line we belong to, runtime */
	0,			/* flags, leave as zero */
	DEVI_CLASS_NONE|DEVI_MODULE_TYPE_DEVICE, /* module type and class, 
						  * a device module usually
						  * doesn't belong to a class
						  */
	"Samp_Dev",		/* name of module, limited to 12 chars
				 * Must match what you specify on the 
				 * cmd line when invoking this module
				 */
	__DATE__,		/* date of compilation, useful in versioning,
				 * output when the -l option is given
				 */
	"i:p:",	        	/* command line parameters for the module,
				 * used by the parm callback, format is that
				 * used in the getopt() call 
				 */
	NULL,			/* private data, initialized in init() cb */
	samp_dev_init,		/* init callback, required */
	samp_dev_reset,		/* reset callback, required */
	NULL,			/* input, device modules don't need this 
				 * as it is invoked by lower modules to pass
				 * data to higher modules 
				 */
	samp_dev_output,	/* used by higher modules to send data to
				 * the device we control 
				 */
        samp_dev_pulse,		/* our pulse callback, used for interrupts, 
				 * and timers
				 */
	samp_dev_parm,		/* parm, required, used to parse any command
				 * line params to the modules
				 */
	samp_dev_devctrl,	/* devctrl, used by higher layer modules to
				 * request non-io services of the device
				 */
	samp_dev_shutdown,	/* shutdown, required, called when module
				 * is being shutdown
				 */
};


/* samp_dev_init
 *
 * This is our init() callback specified in the input_module_t structure.
 * Here we typically allocate space for our private data and assign it
 * to the data member of the module structure.  This callback is called by
 * the input runtime system *before* the module is linked into an event
 * bus line, ie you can't communicate with other modules in your bus line
 * at this point.
 *
 * Return:  0 if ok
 *         -1 on error
 */

static int 
samp_dev_init(input_module_t *module)
{
        struct private_data *dp = module->data;

        if(!module->data) {

                dp = module->data = malloc(sizeof(struct private_data));
                if (!dp) {
                        errno_print("malloc");
                        return (-1);
                }
                
                memset(dp, 0, sizeof(struct private_data));

                dp->irq = DEFAULT_IRQ;
                dp->ioport = DEFAULT_IOPORT;
                dp->irq_pc = DEVI_PULSE_ALLOC;
        }

        return (0);
}

/* samp_dev_parm
 *
 * This callback is called by the input runtime system to parse any
 * command line parameters given to the module.  The argument opt
 * is the command line option given and optarg contains the argument
 * to the option if any.  This callback is called *before* the module
 * is linked into an event bus line.
 *
 * Return  0 if ok
 *        -1 on error
 */

static int 
samp_dev_parm(input_module_t *module, int opt, char *optarg)
{
        struct private_data *dp = module->data;

        switch (opt) {

                /* user specified irq on command line */
        case 'i':
                dp->irq = strtol(optarg, 0, 0);
                break;

                /* user specified ioport on command line */
                
        case 'p':
                dp->ioport = strtoul(optarg, 0, 16);
                break;
                
        default:
                return (-1);
        }
        
        return (0);
}


/* samp_dev_reset
 *
 * The reset callback is called when our module is linked into an
 * event bus line.  Now communication is possible with the other
 * modules in the bus line.  In here we will setup our device for
 * action.  We will map in our IO range, attach to our irq, and setup
 * our timers.  
 *
 * Return   0 if ok
 *         -1 on error
 * 
 */


static int 
samp_dev_reset(input_module_t *module)
{
        struct private_data	*dp   = module->data;
	struct sigevent		ev;

        /* map in device memory */
        
        if (verbosity >= 2)
                printf("mapping in %x\n", (unsigned int) dp->ioport);
        
        /* 
         * Here we use the library call devi_request_iorange() which
         * will call mmap_device_io for us and reserve the io range
         * with Procs resource db manager
         */

        dp->iobase = devi_request_iorange(dp->ioport, 8, DEVI_SHARE_RSRC);
        if (dp->iobase == (uintptr_t) MAP_FAILED) 
                return (-1);

        /* 
         * Now you can access your devices registers with the in*()/out*() calls.
         * Allocate and configure anything else you need like buffers, etc.
         */
    
        /* set up irq handling */
                
        if (verbosity >= 2)
                printf("Attaching to irq %#x\n", dp->irq);

        /* We use the library call devi_register_interrupt() to attach
         * out irq.  This call will setup a pulse to send to the input
         * runtime system when the specified irq triggers.  The pulse will
         * be sent at the priority specified in the second argument (15 in
         * our case) and with the code returned to us in the third argument.
         * Note we initialized the dp->irq_pc member to DEVI_PULSE_ALLOC to
         * tell the devi_register_interrupt() call to allocate a number for
         * pulse itself.
         * When the input runtime system receives the pulse it
         * will call our pulse() callback.
         */

        dp->iid = devi_register_interrupt(dp->irq, 15, &dp->irq_pc, module, 
                                          NULL, 0);
        if (dp->iid < 0) {
                fprintf(stderr, "Unable to register interrupt %d\n", dp->irq);
                return (-1);
        }

        /* Note nothing is preventing you from not using these calls.  If
         * you want to have an actual interrupt handler, feel free to
         * call InterruptAttach() and assign it a handler.  
         */

        /* Lets also set up a 5 second timer.  We can do this by first
         * calling the library call devi_register_pulse.  This call
         * will associate a pulse code with a function.  When the input
         * runtime systems receives a pulse with the code returned by
         * devi_register_pulse it will call the associated function.  By
         * passing in DEVI_PULSE_ALLOC into the second argument we are asking
         * the system to allocate a pulse code for us.  By passing in NULL for
         * the third argument, we are telling the system to associate the 
         * pulse() member of the input module passed in as the first arg.
         */

        dp->tim_pc = devi_register_pulse(module, DEVI_PULSE_ALLOC, NULL);
        if (dp->tim_pc < 0) {
                fprintf(stderr, "Unable to register pulse\n");
                return (-1);
        }
        
        /* now actually setup the timer */

        memset(&ev, 0, sizeof(ev));
        SIGEV_PULSE_INIT(&ev, devi_get_coid(), 15, dp->tim_pc, 0);

        if (timer_create(CLOCK_REALTIME, &ev, &dp->timer_id) < 0) {
                errno_print("timer_create");
                return (-1);
        }

        dp->ts.it_value.tv_sec = 5;
        dp->ts.it_interval.tv_sec = 5;

        if (timer_settime(dp->timer_id, 0, &dp->ts, NULL) < 0) {
                errno_print("timer_settime");
                return (-1);
        }

        return (0);
}

/* samp_dev_pulse
 *
 * This is our callback for event notifications from the input runtime
 * system.  In our case we are using it for interrupt and timer handling
 *
 * Return 0.
 */

static int 
samp_dev_pulse(message_context_t *ctp, int code, unsigned flags, void *data)
{
        input_module_t		*module = (input_module_t *) data;
        struct private_data	*dp = module->data;
        input_module_t 		*up = module->up;
        int 			count = 0;
	char 			*buffer = NULL;

        /* Check for irq */
        if (code == dp->irq_pc) {

                /* process irq, read in data etc */

                /* now if you have data to pass up to the next layer
                 * you pass it via the input() callback of the module
                 * above you like this:
                 */
                 (up->input)(up, count, buffer);

                /* remember to unmask the irq! */
                InterruptUnmask(dp->irq, -1);

        }

        else if (code == dp->tim_pc) {

                /* process the timer */

        }

        return (0);
}


/* samp_dev_output
 *
 * This callback is usually called by higher layer modules asking
 * for data to be sent to the device.
 *
 * Return  0 if ok
 *        -1 on error
 */

static int 
samp_dev_output(input_module_t *module, void *data, int count)
{
        struct private_data *dp = module->data;
        /* send output to device */
        dp = dp;

        return (0);
}
    

/* samp_dev_devctrl
 *
 * The devctrl() callback is used by modules in an event bus line to
 * send commands to each other.  The current list of devctrl events
 * are listed in the header const.h in the include directory.
 * The third argument ptr is a pointer to any data associated with
 * the event.  These data types are listed in the file devi.h 
 * and prefixed with devctl, eg struct devctl_led is used when sending
 * a command to the keyboard device modules asking it to turn on/off its
 * lights.
 *
 * Return  0 if ok
 *        -1 on error
 */

static int 
samp_dev_devctrl(input_module_t *module, int event, void *ptr)
{
        struct private_data *dp = module->data;

        dp = dp;
        switch(event) {
/*
        case DEVCTL_EXAMPLE:
                return (0);
*/
	  
        default:
                return (-1);
        }

        return (EOK);
}


/* samp_dev_shutdown
 *
 * This callback is called when the input runtime system is shutting down.
 * Perform any device specific shutdown here.
 *
 * Return  0 if ok
 *        -1 on error
 */

static int 
samp_dev_shutdown(input_module_t *module, int ms)
{
        struct private_data *dp = module->data;

        /* perform any device specific shutdown */

        dp = dp;
        return (0);
}





