/*
 * Copyright 2001, QNX Software Systems Ltd. All Rights Reserved
 *
 * This source code has been published by QNX Software Systems Ltd. (QSSL).
 * However, any use, reproduction, modification, distribution or transfer of
 * this software, or any software which includes or is based upon any of this
 * code, is only permitted under the terms of the QNX Open Community License
 * version 1.0 (see licensing.qnx.com for details) or as otherwise expressly
 * authorized by a written license agreement from QSSL. For more information,
 * please email licensing@qnx.com.
 */

/*
 * samp_proto
 *
 * A heavily commented version of the microsoft mouse protocol.  
 * The MS mouse protocol is a byte oriented protocol used by
 * serial mice to communicate relative position coordinates
 * to the system.  
 *
 * The serial parameters typically required by the mouse are
 * 7 bits, 1200 baud, no parity.  The data sent by the mouse
 * has the following default format:
 * 
 * 
 *              6  5  4  3  2  1  0
 * Byte 1:      1  L  R  Y7 Y6 X7 X6
 * Byte 2:      0  X5 X4 X3 X2 X1 X0
 * Byte 3:      0  Y5 Y4 Y3 Y2 Y1 Y0
 * 
 * L - 1 if left button pressed
 * R - 1 if right button pressed
 *
 * Y7-Y0 : Y coordinate
 * X7-X0 : X coordinate
 *
 * Note this module also knows how to talk to IntelliMice (wheel mice)
 * and 3 button mice.  These mice will send in a fourth byte indicating
 * wheel movement and middle button status.  Command line parameters to
 * the module will indicate whether or not processing of the fourth byte
 * is necessary.
 */
#include "devi.h"

#define FLAG_INIT		0x0100
#define FLAG_3BUTTON		0x0200
#define FLAG_MIDDLE		0x0400
#define FLAG_RESET		0x0800
#define FLAG_INTELLIMSE		0x1000

static int samp_proto_init(input_module_t *module);
static int samp_proto_devctrl(input_module_t *module, int event, void *ptr);
static int samp_proto_reset(input_module_t *module);
static int samp_proto_input(input_module_t *module, int num, void *arg);
static int samp_proto_parm(input_module_t *module,int opt,char *optarg);
static int samp_proto_shutdown(input_module_t *module, int delay);

/* Our protocol module is represented by the following input_module_t
 * data structure.  We always create one static instance of the module.
 * If more than one are needed, ie in multiple bus lines; then the system
 * will allocate a new module and copy the contents of the static one
 * into it.
 */

input_module_t	samp_proto = {
	NULL,			/* up, filled in at runtime */
	NULL,			/* down, filled in at runtime */
	NULL,			/* line we belong to, filled in at runtime */
	0,			/* flags, leave as zero */
	DEVI_CLASS_REL|DEVI_MODULE_TYPE_PROTO, /* Our type, we are a 
						* protocol module or class
						* relative.  This info will
						* tell the runtime system
						* which filter module to
						* link above us
						*/
	"Samp_Proto",		/* our name, must match what you specify
				 * on the command line when invoking this
				 * module
				 */
	__DATE__,		/* date of compilation, used for output when
				 * the cmdline option -l is given to the
				 * driver
				 */
	"iR3b:",		/* command line parameters */
	NULL,			/* pointer to private data, set this up
				 * in the init() callback 
				 */
	samp_proto_init,	/* our init() callback, required */
	samp_proto_reset,	/* our reset() callback, required */
	samp_proto_input,	/* our input() callback */
	NULL,			/* output(), not used */
	NULL,			/* pulse(), not used */
	samp_proto_parm,	/* our parm() callback, required */
	samp_proto_devctrl,	/* our devctrl() callback */
	samp_proto_shutdown	/* our shutdown() callback, required */
};

/* Our modules private data */

struct private_data {
	int			flags; 
	struct packet_rel	mp;
	int			baud;
	int			state;
	int			packet_size;
};


/* samp_proto_init
 *
 * This is our init() callback specified in the input_module_t structure.
 * Here we typically allocate space for our private data and assign it
 * to the 'data' member of the module structure.  This callback is called by
 * the input runtime system *before* the module is linked into an event
 * bus line, ie you can't communicate with other modules in your bus line
 * at this point.
 *
 * Return:  0 if ok
 *         -1 on error
 */

static int 
samp_proto_init(input_module_t *module)
{
	struct private_data *dp = module->data;

	if(!module->data)  {
		if(!(dp = module->data = scalloc(sizeof *dp))) {
			return (-1);
		}
		dp->baud = 1200;

		/* default is to reset mouse */
		dp->flags = FLAG_RESET;
		dp->packet_size = 3;
	}
	return (0);
}


/* samp_proto_parm
 *
 * This callback is called by the input runtime system to parse any
 * command line parameters given to the module.  The argument opt
 * is the command line option given and optarg contains the argument
 * to the option if any.  This callback is called *before* the module
 * is linked into an event bus line.
 *
 * Return  0 if ok
 *        -1 on error
 */

static int 
samp_proto_parm(input_module_t *module,int opt,char *optarg)
{
	struct private_data *dp = module->data;

 	switch (opt) {

		case 'i':	/* wheel mouse */
			dp->packet_size = 4; /* wheel mouse sends 4 bytes */
			dp->flags |= FLAG_INTELLIMSE;
			break;
			
		case 'R':	/* don't reset the device */
			dp->flags &= ~FLAG_RESET;
			break;

 		case '3':	/* 3 button MS mouse */
			dp->flags |= FLAG_3BUTTON;
 			break;

 		case 'b':	/* baud rate to device */
			dp->baud = strtol(optarg,0,0);
 			break;

 		default:
 			return (-1);
 	}

	return (0);
}


/* samp_proto_reset
 *
 * The reset callback is called when our module is linked into an
 * event bus line.  Now communication is possible with the other
 * modules in the bus line.  We'll need to tell our device how
 * to setup the comm parameters, ie baud rate and data format.  To
 * do this we'll invoke the devctrl() callback of our device layer
 * module, accessible via the down pointer in our input_module_t
 * structure.
 *
 * Return   0 if ok
 *         -1 on error
 * 
 */

static int 
samp_proto_reset(input_module_t *module)
{
	struct devctl_setbaud	devctl;
	struct devctl_settrig	trig;
	struct private_data	*dp = module->data;
	input_module_t		*down = module->down;
	
	if((dp->flags & FLAG_INIT) == 0) {

		devctl.baud = dp->baud;
		devctl.cflag = PNONE | STOP2 | CS7;

                /* Now send commands to our device layer module, note that
                 * in some cases we don't check the return value from the
                 * devctrl.  This is b/c some device layer mods don't support
                 * all the commands we send. 
                 */

		if(down && down->devctrl) {

			/* reset the device module if required */
			if(dp->flags & FLAG_RESET) {
				(down->devctrl)(down, DEVCTL_RESET, NULL);
			}

			/* send the baud/format devctrl to our dev module */
			if ((down->devctrl)(down, DEVCTL_SETBAUD, &devctl) < 0) {
                                fprintf(stderr, "Unable to set comm parameters to device\n");
                                return (-1);
                        }

			/* The trigger level tells the device module how
			 * many bytes to wait for before sending us any
			 * data.  We'll initially set it to our packet
			 * size.
			 */
			
			trig.trigger_level = dp->packet_size;
			(down->devctrl)(down, DEVCTL_SETTRIG, &trig);

		}	
                else return (-1);

		dp->flags |= FLAG_INIT;
	}
	return (0);
}


/* samp_proto_devctrl
 *
 * The devctrl() callback is used by modules in an event bus line to
 * send commands to each other.  The current list of devctrl events
 * are listed in the header const.h in the include directory.
 * The third argument ptr is a pointer to any data associated with
 * the event.  These data types are listed in the file devi.h 
 * and prefixed with devctl, eg struct devctl_led is used when sending
 * a command to the keyboard device modules asking it to turn on/off its
 * lights.
 *
 * Return  0 if ok
 *        -1 on error
 */


static int 
samp_proto_devctrl(input_module_t *module, int event, void *ptr)
{
	struct private_data 	*dp = module->data;
	input_module_t		*down = module->down;
	
	switch(event) {

	case	DEVCTL_GETDEVFLAGS:
		*(unsigned short *)ptr=(dp->flags & FLAGS_GLOBAL);
		break;

	case	DEVCTL_GETPTRBTNS:
		if(dp->flags & FLAG_3BUTTON) {
			*(unsigned long *)ptr=7L;
		} else {
			*(unsigned long *)ptr=5L;
		}
		break;

	case	DEVCTL_GETPTRCOORD:
		*(unsigned char *)ptr='\02';
		break;

	default:                /* if we don't understand the cmd, pass
                                 * it on down to the next module 
                                 */
		if(down && down->devctrl) {
			return((down->devctrl)(down, event, ptr));
		}
		return (-1);
	}
	
	return (0);
}


/* samp_proto_input
 *
 * This is our main protocol processing function.  It will be called
 * by the device layer to pass us data to process.  Our job is to
 * interpret the data according to the MS mouse protocol, create a
 * packet_rel data structure, fill it in and hand it off to the
 * the filter layer module above us. The protocol is processed using
 * simple state machine.  State info is kept in our module's private
 * data.
 * 
 * Returns:  0 on success
 *          -1 on error
 */

static int 
samp_proto_input(input_module_t *module, int num, void *arg)
{
        int			        i, enqueue = 0;
        struct private_data		*dp = module->data;
        input_module_t			*down = module->down;
        input_module_t			*up = module->up;
        struct devctl_settrig		trig;
        char 				*p = (char *) arg;
        
        if(dp->flags & FLAG_INIT) {
                
                for(i = 0; i < num; i++, p++) {
                        
                        if(!(dp->flags & FLAG_INTELLIMSE)) {
                                
                                if(*p & 0x40) {
                                        if(!(dp->flags & FLAG_3BUTTON)) {
                                                dp->flags &= ~FLAG_MIDDLE;
                                        }
                                        dp->state = 0;
                                        
                                } else if (dp->state == 0) {
                                        if((*p & 0x5C) == 0) {
                                                if(*p & 0x20) {
                                                        dp->mp.buttons |= _POINTER_BUTTON_MIDDLE;
                                                        dp->flags |= FLAG_MIDDLE;
                                                } else {
                                                        dp->flags &= ~FLAG_MIDDLE;
                                                }
                                                enqueue = 1;
                                                
                                        } else {
                                                continue;			/* framing error */
                                        }
                                }
			}
			else if (*p & 0x40) /* sync up the intellimouse */
				dp->state = 0;
                        
                        switch(dp->state) {
                                
                        case 0:
                                if(enqueue) {
                                        break;
                                }
                                dp->mp.buttons = ((*p & 0x20) ? _POINTER_BUTTON_LEFT : 0) | 
                                        ((*p & 0x10) ? _POINTER_BUTTON_RIGHT : 0);
                                
                                dp->mp.dy = (*p & 0x0c) << 4;
                                dp->mp.dx = (*p & 0x03) << 6;
                                dp->state++;
                                break;
                                
                        case 1:
                                dp->mp.dx |= (*p & 0x3f);
                                dp->state++;
                                break;
                                
                        case 2:
                                dp->mp.dy |= (*p & 0x3f);
                                if(dp->mp.dy & 0x80) {
                                        dp->mp.dy |= 0xff00;
                                }
                                if(dp->mp.dx & 0x80) {
                                        dp->mp.dx |= 0xff00;
                                }
                                if(dp->flags & FLAG_INTELLIMSE) {
                                        dp->mp.flags |= PTR_Z_DATA;
                                        dp->state = 3;
                                } else {
                                        dp->mp.flags &= ~PTR_Z_DATA;
                                        if(!(dp->flags & FLAG_MIDDLE)) {
                                                enqueue = 1;
                                        }
                                        dp->state=0;
                                }
                                break;
                                
                        case 3:
                                dp->mp.dz = (*p & 0x01);
                                if(*p & 0x08) {
                                        dp->mp.dz = -dp->mp.dz;
                                }
                                if(*p & 0xf0) {
                                        dp->mp.buttons |= _POINTER_BUTTON_MIDDLE;
                                } else {
                                        dp->mp.buttons &= ~_POINTER_BUTTON_MIDDLE;
                                }
                                enqueue = 1;
                                break;
                        }
                        
                        /* send our complete packet_rel up to the filter layer */
                        if(enqueue) {
                                dp->mp.dy = -(dp->mp.dy);
                                clk_get(&dp->mp.timestamp);
                                (up->input)(up, 1, &dp->mp);
                                enqueue = 0;
                                dp->state = 0;
                        }
                }
                
                if(dp->flags & FLAG_3BUTTON) {
                        trig.trigger_level = 1;
                        (down->devctrl)(down, DEVCTL_SETTRIG, &trig);
                } else {
                        
                        trig.trigger_level = (dp->flags & FLAG_MIDDLE) ? 4 : 
                                dp->packet_size - dp->state;
                        
                        (down->devctrl)(down, DEVCTL_SETTRIG, &trig);
                }
        }
	
        return (0);
}


/* samp_proto_shutdown
 *
 * This callback is called when the input runtime system is shutting down.
 *
 * Return  0 if ok
 *        -1 on error
 */

static int 
samp_proto_shutdown(input_module_t *module, int delay)
{
	return (0);
}
